from dataclasses import dataclass
from antlr4.error.ErrorListener import ConsoleErrorListener, ErrorListener


class NewErrorListener(ConsoleErrorListener):
    INSTANCE = None

    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        raise SyntaxException("Error on line " + str(line) +
                              " col " + str(column) + ": " + offendingSymbol.text)


NewErrorListener.INSTANCE = NewErrorListener()


class SyntaxException(Exception):
    def __init__(self, msg):
        self.message = msg


class UnclosedComment(Exception):
    """Error raised when a comment is not closed."""

    def __init__(self, line: int, column: int):
        self.line = line
        self.column = column

    def __str__(self):
        return f"Unclosed comment at line {self.line}, column {self.column}."


class StaticError(Exception):
    """Base class for static errors."""
    pass


class NoMainFoundInProject(StaticError):
    """Error raised when no main function is found in the project."""

    def __str__(self):
        return "No main function found in the project."


@dataclass
class IncludeNotFound(StaticError):
    """Error raised when an include file is not found."""
    filename: str

    def __str__(self):
        return f"Include file '{self.filename}' not found."


class MultipleMainFound(StaticError):
    """Error raised when multiple main functions are found in the project."""

    def __str__(self):
        return "Multiple main functions found in the project."



from CDG import *
from Report import *
from StaticCheck import SignalType
from AST import *
import sys
from tqdm import tqdm

sys.setrecursionlimit(5000)


class Detector:
    def __init__(self, graphs):
        self.graphs = graphs
        self.reports = {}
        self.as_const = {}

    def detect(self):
        for graph in self.graphs.values():
            self.reports[graph.name] = {}
            print(
                f"[Info]       Starting the analysis process of graph {graph.name}.")

            print("[Info]       Detecting unconstrainted output...")
            self.detect_unconstrainted_output(graph)

            print("[Info]       Detecting unconstrained component input...")
            self.detect_unconstrained_comp_input(graph)

            print("[Info]       Detecting data flow constraint discrepancy...")
            self.detect_data_flow_constraint_discrepancy(graph)

            print("[Info]       Detecting unused component output...")
            self.detect_unused_comp_output(graph)

            print("[Info]       Detecting type mismatch...")
            self.detect_type_mismatch(graph)

            print("[Info]       Detecting assignment misuse...")
            self.detect_assignment_misue(graph)

            print("[Info]       Detecting unused signals...")
            self.detect_unused_signal(graph)

            print("[Info]       Detecting divide by zero unsafe...")
            self.detect_divide_by_zero_unsafe(graph)

            print("[Info]       Detecting nondeterministic data flow...")
            self.detect_nondeterministic_data_flow(graph)
        return self.reports

    def detect_unconstrainted_output(self, graph):
        results = []
        for n_id in tqdm(graph.components[graph.name][SignalType.OUTPUT]):
            node = graph.nodes[n_id]
            if self.unconstrainted_ouput(graph, node):
                results.append(Report(ReportType.WARNING, node.locate,
                               f"Output signal '{node.id}' is not constrained by any constraint."))
        self.reports[graph.name]["unconstrained_output"] = results

    def unconstrainted_ouput(self, graph, node):
        if not node.is_signal_out():
            return False
        constrainted_by_input = self.constrainted_by_input(graph, node)
        constrainted_as_const = self.constrainted_as_const(graph, node)
        return not (constrainted_by_input or constrainted_as_const)

    def constrainted_by_input(self, graph, node):
        for n1 in graph.nodes.values():
            if not n1.is_signal_in():
                continue
            if self.is_constrainted(graph, node, n1):
                return True
        return False

    def constrainted_as_const(self, graph, node):
        if node.id in self.as_const:
            return self.as_const[node.id]
        else:
            self.as_const[node.id] = False
        for edge in node.flow_to:
            if edge.edge_type == EdgeType.CONSTRAINT:
                if edge.node_to.node_type == NodeType.CONSTANT:
                    self.as_const[node.id] = True
                    return True
        for edge in node.flow_from:
            if edge.edge_type == EdgeType.CONSTRAINT:
                node_from = edge.node_from
                if node_from.node_type == NodeType.CONSTANT:
                    self.as_const[node.id] = True
                    return True
                if node_from.node_type == NodeType.SIGNAL and node_from.signal_type == SignalType.INTERMEDIATE:
                    if self.constrainted_as_const(graph, node_from):
                        self.as_const[node.id] = True
                        return True
        self.as_const[node.id] = False
        return False

    def is_constrainted(self, graph, node_a, node_b):
        return graph.has_path_constraint(node_a, node_b)

    def is_depended(self, graph, node_a, node_b):
        if node_a.id not in graph.node_flows_to:
            return False
        return node_b.id in graph.node_flows_to[node_a.id]

    def unconstrained_comp_input(self, graph, node):
        component = node.component.split("|")[0]
        if graph.name == component or not node.is_signal_in():
            return False
        for edge in node.flow_from:
            if edge.edge_type == EdgeType.CONSTRAINT:
                node_from = edge.node_from
                node_from_var_name = node_from.id.split(".")[0]
                node_var_name = node.id.split(".")[0]
                if node_from_var_name != node_var_name:
                    return False
        for edge in node.flow_to:
            if edge.edge_type == EdgeType.CONSTRAINT:
                node_to = edge.node_to
                node_to_var_name = node_to.id.split(".")[0]
                node_var_name = node.id.split(".")[0]
                if node_to_var_name != node_var_name:
                    return False
            if edge.node_to.node_type == NodeType.CONSTANT:
                node_to = edge.node_to
                for e1 in node_to.flow_to:
                    if e1.edge_type == EdgeType.CONSTRAINT:
                        return False
                for e1 in node_to.flow_from:
                    if e1.edge_type == EdgeType.CONSTRAINT:
                        return False
        return True

    def detect_unconstrained_comp_input(self, graph):
        results = []
        for node in tqdm(graph.nodes.values()):
            if self.unconstrained_comp_input(graph, node):
                results.append(Report(ReportType.WARNING, node.locate,
                               f"Input signal '{node.id}' is unconstrained and may accept unchecked values."))
        self.reports[graph.name]["unconstrained component input"] = results

    def detect_data_flow_constraint_discrepancy(self, graph):
        resutlts = []
        for n_id, n_set in tqdm(graph.node_flows_to.items()):
            for n1_id in n_set:
                node = graph.nodes[n_id]
                node_1 = graph.nodes[n1_id]
                if not self.is_constrainted(graph, node, node_1):
                    resutlts.append(Report(ReportType.WARNING, node_1.locate,
                                    f"Signal '{node_1.id}' depends on '{node.id}' via dataflow, but there is no corresponding constraint dependency."))
        self.reports[graph.name]["data flow constraint discrepancy"] = resutlts

    def is_checking_signal(self, node):
        for edge in node.flow_to:
            if edge.edge_type == EdgeType.DEPEND:
                return True
        return False

    def unused_comp_output(self, graph, node):
        component = node.component.split("|")[0]
        if not node.is_signal_out() or component == graph.name:
            return False
        sub_graph = self.graphs[component]
        node_var_name, signal_name = node.id.split(".")
        if signal_name not in sub_graph.nodes:
            return False
        sub_o_node = sub_graph.nodes[signal_name]
        if self.is_checking_signal(sub_o_node):
            return False
        for edge in node.flow_to:
            node_to_var_name = edge.node_to.id.split(".")[0]
            if node_var_name != node_to_var_name:
                return False
        for edge in node.flow_from:
            node_from_var_name = edge.node_from.id.split(".")[0]
            if node_var_name != node_from_var_name:
                return False
        return True

    def detect_unused_comp_output(self, graph):
        resuluts = []
        for node in tqdm(graph.nodes.values()):
            if self.unused_comp_output(graph, node):
                resuluts.append(Report(ReportType.WARNING, node.locate,
                                f"This output '{node.id}' is not checked nor used from the call site."))
        self.reports[graph.name]["unused component output"] = resuluts

    def unsused_signal(self, graph, node):
        component = node.component.split("|")[0]
        if node.is_signal_out() and component == graph.name:
            return False
        if node.node_type == NodeType.CONSTANT:
            return False
        return (len(node.flow_from) + len(node.flow_to)) == 0

    def detect_unused_signal(self, graph):
        results = []
        for node in tqdm(graph.nodes.values()):
            if self.unsused_signal(graph, node):
                results.append(Report(ReportType.WARNING, node.locate,
                               f"This signal '{node.id}' is declared but never used in any computation or constraint."))
        self.reports[graph.name]["unused signal"] = results

    def detect_type_mismatch(self, graph):
        num2bits_required = {"LessThan", "LessEqThan",
                             "GreaterThan", "GreaterEqThan", "BigLessThan"}
        num2bits_like = {"Num2Bits", "Num2Bits_strict",
                         "RangeProof", "MultiRangeProof", "RangeCheck2D"}
        results = []
        for node in tqdm(graph.nodes.values()):
            component = node.component.split("|")[0]
            if not node.is_signal_in() or component == graph.name:
                continue
            template_name = component.split("@")[0]
            if template_name not in num2bits_required:
                continue
            input_nodes = []
            for edge in node.flow_from:
                if edge.edge_type == EdgeType.CONSTRAINT:
                    continue
                if edge.node_from.is_signal():
                    input_nodes.append(edge.node_from)
                if edge.node_from.node_type == NodeType.CONSTANT:
                    for e1 in edge.node_from.flow_from:
                        if e1.edge_type == EdgeType.DEPEND and e1.node_from.is_signal():
                            input_nodes.append(e1.node_from)
            for n1 in input_nodes:
                is_checked = False
                flows_to = graph.flows_to(n1)
                for n2_id in flows_to:
                    n2 = graph.nodes[n2_id]
                    n2_comp = n2.component.split("|")[0]
                    if not n1.is_signal_in() or n2_comp == graph.name:
                        continue
                    template_name_n2 = n2_comp.split("@")[0]
                    if template_name_n2 in num2bits_like:
                        is_checked = True
                        break
                if not is_checked:
                    results.append(Report(ReportType.WARNING, node.locate,
                                   f"Signal '{n1.id}' flows into '{template_name}' without being properly range-checked."))
        self.reports[graph.name]["type mismatch"] = results

    def is_trivial_instruction(self, ast):
        if isinstance(ast, Variable) or isinstance(ast, Number):
            return True
        elif isinstance(ast, PrefixOp):
            return self.is_trivial_instruction(ast.rhe)
        elif isinstance(ast, InfixOp):
            return self.is_trivial_instruction(ast.lhe) and self.is_trivial_instruction(ast.rhe)
        else:
            return False

    def is_rewritable_assignment(self, edge):
        if edge.edge_type != EdgeType.DEPEND or edge.ast is None:
            return False
        node_from = edge.node_from
        node_to = edge.node_to
        for e1 in node_from.flow_to:
            if e1.node_to.id == node_to.id and e1.edge_type == EdgeType.CONSTRAINT:
                return False
        for e1 in node_from.flow_from:
            if e1.node_from.id == node_to.id and e1.edge_type == EdgeType.CONSTRAINT:
                return False
        if node_to.node_type == NodeType.CONSTANT:
            return False
        if node_from.node_type != NodeType.CONSTANT:
            return True
        return self.is_trivial_instruction(edge.ast.rhe)

    def detect_assignment_misue(self, graph):
        results = []
        for edge in tqdm(graph.edges.values()):
            if self.is_rewritable_assignment(edge):
                op = edge.ast.op
                if op == "<--":
                    instead_op = "<=="
                else:
                    instead_op = "==>"
                results.append(Report(ReportType.WARNING, edge.ast.locate,
                               f"Variable {edge.ast.var} is assigned using {op} instead of {instead_op}."))
        self.reports[graph.name]["assignment missue"] = results

    def flat_expr(self, graph, expr):
        if isinstance(expr, Variable):
            name = expr.name
            for node in graph.nodes.values():
                if name in node.id and node.is_signal():
                    return True
        elif isinstance(expr, InfixOp):
            return self.flat_expr(graph, expr.lhe) or self.flat_expr(graph, expr.rhe)
        elif isinstance(expr, PrefixOp):
            return self.flat_expr(graph, expr.rhe)
        elif isinstance(expr, InlineSwitchOp):
            return self.flat_expr(graph, expr.cond) or self.flat_expr(graph, expr.if_true) or self.flat_expr(graph, expr.if_false)
        elif isinstance(expr, Call):
            for arg in expr.args:
                if self.flat_expr(graph, arg):
                    return True
        return False

    def is_denominator_with_signal(self, graph, expr):
        if isinstance(expr, InfixOp):
            if expr.infix_op == "/":
                if self.flat_expr(graph, expr.rhe):
                    return True
            if self.is_denominator_with_signal(graph, expr.lhe) or self.is_denominator_with_signal(graph, expr.rhe):
                return True
        elif isinstance(expr, InlineSwitchOp):
            return self.is_denominator_with_signal(graph, expr.cond) or self.is_denominator_with_signal(graph, expr.if_true) or self.is_denominator_with_signal(graph, expr.if_false)
        return False

    def detect_divide_by_zero_unsafe(self, graph):
        results = []
        for edge in tqdm(graph.edges.values()):
            if edge.ast and self.is_denominator_with_signal(graph, edge.ast.rhe):
                results.append(Report(ReportType.WARNING, edge.ast.locate,
                               f"Potential divide-by-zero issue detected."))
        self.reports[graph.name]["divide by zero"] = results

    def is_branch_cond_with_signal(self, graph, expr):
        if isinstance(expr, InfixOp):
            if self.is_branch_cond_with_signal(graph, expr.lhe) or self.is_branch_cond_with_signal(graph, expr.rhe):
                return True
        elif isinstance(expr, InlineSwitchOp):
            return self.flat_expr(graph, expr.cond)
        return False

    def detect_nondeterministic_data_flow(self, graph):
        results = []
        for edge in tqdm(graph.edges.values()):
            if edge.ast and self.is_branch_cond_with_signal(graph, edge.ast.rhe):
                results.append(Report(ReportType.WARNING, edge.ast.locate,
                                      "Potential non-deterministic dataflow: conditional assignment depends on a signal."))
        self.reports[graph.name]["nondeterministic data flow"] = results



from CircomVisitor import CircomVisitor
from CircomLexer import CircomLexer
from CircomParser import CircomParser
from AST import *
import Errors
import os
from antlr4 import *
import sys
from Report import Report, ReportType

sys.setrecursionlimit(5000)


class ASTGeneration(CircomVisitor):
    def generate_ast(self, filename, tree, path_env):
        print("[Info]       Generating AST for:", filename)
        self.file_name = filename
        ast_tree = self.visit(tree)
        base_dir = os.path.dirname(filename)
        for include in ast_tree.includes:
            raw_include_path = include.path.strip('"').strip("'")
            absolute_path = os.path.abspath(
                os.path.join(base_dir, raw_include_path))
            if not os.path.isfile(absolute_path):
                Report(ReportType.ERROR, ast_tree.locate,
                       Errors.IncludeNotFound(absolute_path)).show()
                return None
            if absolute_path in path_env:
                continue
            path_env.append(absolute_path)
            inputstream = FileStream(absolute_path, encoding='utf-8')
            lexer = CircomLexer(inputstream)
            listener = Errors.NewErrorListener.INSTANCE
            tokens = CommonTokenStream(lexer)
            parser = CircomParser(tokens)
            parser.removeErrorListeners()
            parser.addErrorListener(listener)
            try:
                tree = parser.program()
            except Errors.SyntaxException as f:
                Report(ReportType.ERROR, FileLocation(
                    absolute_path, None, None), f.message, False).show()
                return None
            except Exception as e:
                Report(ReportType.ERROR, FileLocation(
                    absolute_path, None, None), str(e), False).show()
                return None
            try:
                asttree = ASTGeneration().generate_ast(absolute_path, tree, path_env)
                if asttree.main_component:
                    Report(ReportType.ERROR, FileLocation(
                        absolute_path, None, None), "Main component found in include file", False).show()
                    return None
                ast_tree.definitions += asttree.definitions
            except Exception as e:
                Report(ReportType.ERROR, FileLocation(
                    absolute_path, None, None), str(e), False).show()
                return None
        return ast_tree

    # program: pragma_definition? custom_gate? include_list definition_list main_option EOF;
    def visitProgram(self, ctx: CircomParser.ProgramContext):
        compile_version = self.visit(
            ctx.pragma_definition()) if ctx.pragma_definition() else None
        custom_gates = True if ctx.custom_gate() else False
        definitions = self.visit(ctx.definition_list())
        custom_gates_declare = False
        includes = self.visit(ctx.include_list())
        main_component = self.visit(ctx.main_option())
        for definition in definitions:
            if isinstance(definition, Template) and definition.is_custom_gate:
                custom_gates_declare = True
                break
        return Program(FileLocation(self.file_name, ctx.start, ctx.stop), compile_version, custom_gates, custom_gates_declare, includes, definitions, main_component)

    # version: NUMBER '.' NUMBER '.' NUMBER ;
    def visitVersion(self, ctx: CircomParser.VersionContext):
        return (int(ctx.NUMBER(0).getText()), int(ctx.NUMBER(1).getText()), int(ctx.NUMBER(2).getText()))

    # main_option: main_component | ;
    def visitMain_option(self, ctx: CircomParser.Main_optionContext):
        return self.visit(ctx.main_component()) if ctx.main_component() else None

    # definition_block: function_definition | template_definition;
    def visitDefinition_block(self, ctx: CircomParser.Definition_blockContext):
        return self.visit(ctx.getChild(0))

    # definition_list: definition_block definition_list | ;
    def visitDefinition_list(self, ctx: CircomParser.Definition_listContext):
        return [self.visit(ctx.definition_block())] + self.visit(ctx.definition_list()) if ctx.definition_block() else []

    # pragma_definition: PRAGMA CIRCOM version SEMICOLON;
    def visitPragma_definition(self, ctx: CircomParser.Pragma_definitionContext):
        return self.visit(ctx.version())

    # custom_gate: PRAGMA CUSTOM_TEMPLATES SEMICOLON;
    def visitCustom_gate(self, ctx: CircomParser.Custom_gateContext):
        pass

    # include_list: include_definition include_list | ;
    def visitInclude_list(self, ctx: CircomParser.Include_listContext):
        return [self.visit(ctx.include_definition())] + self.visit(ctx.include_list()) if ctx.include_definition() else []

    # include_definition: INCLUDE STRING SEMICOLON;
    def visitInclude_definition(self, ctx: CircomParser.Include_definitionContext):
        return Include(FileLocation(self.file_name, ctx.start, ctx.stop), ctx.STRING().getText())

    # public_list: LC PUBLIC LB identifier_list RB RC | ;
    def visitPublic_list(self, ctx: CircomParser.Public_listContext):
        return self.visit(ctx.identifier_list()) if ctx.identifier_list() else []

    # main_component: COMPONENT MAIN public_list ASSIGNMENT expression SEMICOLON;
    def visitMain_component(self, ctx: CircomParser.Main_componentContext):
        return MainComponent(FileLocation(self.file_name, ctx.start, ctx.stop), self.visit(ctx.public_list()), self.visit(ctx.expression()))

    # function_definition: FUNCTION IDENTIFIER LP identifier_list_option RP block;
    def visitFunction_definition(self, ctx: CircomParser.Function_definitionContext):
        return Function(FileLocation(self.file_name, ctx.start, ctx.stop), ctx.IDENTIFIER().getText(), self.visit(ctx.identifier_list_option()), self.visit(ctx.block()))

    # template_definition: TEMPLATE custom_option parallel_option IDENTIFIER LP identifier_list_option RP block;
    def visitTemplate_definition(self, ctx: CircomParser.Template_definitionContext):
        is_custom = self.visit(ctx.custom_option())
        is_parallel = self.visit(ctx.parallel_option())
        return Template(FileLocation(self.file_name, ctx.start, ctx.stop), ctx.IDENTIFIER().getText(), self.visit(ctx.identifier_list_option()), self.visit(ctx.block()), is_parallel, is_custom)

    # identifier_list_option: identifier_list | ;
    def visitIdentifier_list_option(self, ctx: CircomParser.Identifier_list_optionContext):
        return self.visit(ctx.identifier_list()) if ctx.identifier_list() else []

    # custom_option: CUSTOM | ;
    def visitCustom_option(self, ctx: CircomParser.Custom_optionContext):
        return ctx.CUSTOM().getText() if ctx.CUSTOM() else None

    # parallel_option: PARALLEL | ;
    def visitParallel_option(self, ctx: CircomParser.Parallel_optionContext):
        return ctx.PARALLEL().getText() if ctx.PARALLEL() else None

    # block: LC statement_list RC;
    def visitBlock(self, ctx: CircomParser.BlockContext):
        return Block(FileLocation(self.file_name, ctx.start, ctx.stop), self.visit(ctx.statement_list()))

    # statement
    #     : declaration_statement
    #     | if_statement
    #     | regular_statement
    #     ;
    def visitStatement(self, ctx: CircomParser.StatementContext):
        return self.visit(ctx.getChild(0))

    # statement_list: statement statement_list | ;
    def visitStatement_list(self, ctx: CircomParser.Statement_listContext):
        return [self.visit(ctx.statement())] + self.visit(ctx.statement_list()) if ctx.statement() else []

    # declaration_statement: declaration SEMICOLON;
    def visitDeclaration_statement(self, ctx):
        return self.visit(ctx.declaration())

    # expression_statement: expression SEMICOLON;
    def visitExpression_statement(self, ctx: CircomParser.Expression_statementContext):
        return MultiSubstitution(FileLocation(self.file_name, ctx.start, ctx.stop), TupleExpr(FileLocation(self.file_name, ctx.start, ctx.stop), []), "<==", self.visit(ctx.expression()))

    # substitutions
    #     : expression assign_opcode expression
    #     | expression RIGHT_ASSIGNMENT expression
    #     | expression RIGHT_CONSTRAINT expression
    #     | variable ASSIGNMENT_WITH_OP expression
    #     | variable SELF_OP
    #     ;
    def visitSubstitutions(self, ctx: CircomParser.SubstitutionsContext):
        if ctx.SELF_OP():
            op = ctx.SELF_OP().getText()[0]
            variable = self.visit(ctx.variable())
            return Substitution(FileLocation(self.file_name, ctx.start, ctx.stop), variable.name, variable.access, "=", InfixOp(FileLocation(self.file_name, ctx.start, ctx.stop), variable, op, Number(FileLocation(self.file_name, ctx.start, ctx.stop), 1)))
        elif ctx.ASSIGNMENT_WITH_OP():
            op = ctx.ASSIGNMENT_WITH_OP().getText()[:-1]
            variable = self.visit(ctx.variable())
            return Substitution(FileLocation(self.file_name, ctx.start, ctx.stop), variable.name, variable.access, "=", InfixOp(FileLocation(self.file_name, ctx.start, ctx.stop), variable, op, self.visit(ctx.expression(0))))
        elif ctx.RIGHT_ASSIGNMENT() or ctx.RIGHT_CONSTRAINT():
            if ctx.RIGHT_ASSIGNMENT():
                op = ctx.RIGHT_ASSIGNMENT().getText()
            else:
                op = ctx.RIGHT_CONSTRAINT().getText()
            var = self.visit(ctx.expression(1))
            return Substitution(FileLocation(self.file_name, ctx.start, ctx.stop), var.name, var.access, op, self.visit(ctx.expression(0))) if isinstance(var, Variable) else MultiSubstitution(FileLocation(self.file_name, ctx.start, ctx.stop), var, op, self.visit(ctx.expression(0)))
        else:
            op = self.visit(ctx.assign_opcode())
            var = self.visit(ctx.expression(0))
            return Substitution(FileLocation(self.file_name, ctx.start, ctx.stop), var.name, var.access, op, self.visit(ctx.expression(1))) if isinstance(var, Variable) else MultiSubstitution(FileLocation(self.file_name, ctx.start, ctx.stop), var, op, self.visit(ctx.expression(1)))

    # substitutions_statement: substitutions SEMICOLON;
    def visitSubstitutions_statement(self, ctx: CircomParser.Substitutions_statementContext):
        return self.visit(ctx.getChild(0))

    # if_statement
    #     : IF LP expression RP if_statement
    #     | IF LP expression RP regular_statement
    #     | IF LP expression RP regular_statement ELSE if_statement
    #     | IF LP expression RP regular_statement ELSE regular_statement
    #     ;
    def visitIf_statement(self, ctx: CircomParser.If_statementContext):
        if ctx.ELSE():
            return IfThenElse(FileLocation(self.file_name, ctx.start, ctx.stop), self.visit(ctx.expression()), self.visit(ctx.regular_statement(0)), self.visit(ctx.if_statement()) if ctx.if_statement() else self.visit(ctx.regular_statement(1)))
        else:
            return IfThenElse(FileLocation(self.file_name, ctx.start, ctx.stop), self.visit(ctx.expression()), self.visit(ctx.if_statement()) if ctx.if_statement() else self.visit(ctx.regular_statement(0)), None)

    # regular_statement
    #     : block
    #     | expression_statement
    #     | substitutions_statement
    #     | for_statement
    #     | while_statement
    #     | equal_constraint_statement
    #     | return_statement
    #     | assert_statement
    #     | log_statement
    #     ;
    def visitRegular_statement(self, ctx: CircomParser.Regular_statementContext):
        return self.visit(ctx.getChild(0))

    # for_statement
    #     : FOR LP declaration SEMICOLON expression SEMICOLON substitutions RP regular_statement
    #     | FOR LP substitutions SEMICOLON expression SEMICOLON substitutions RP regular_statement
    #     ;
    def visitFor_statement(self, ctx: CircomParser.For_statementContext):
        if ctx.declaration():
            init = self.visit(ctx.declaration())
            while_body = Block(FileLocation(self.file_name, ctx.start, ctx.stop), [
                               self.visit(ctx.regular_statement()), self.visit(ctx.substitutions(0))])
        else:
            init = self.visit(ctx.substitutions(0))
            while_body = Block(FileLocation(self.file_name, ctx.start, ctx.stop), [
                               self.visit(ctx.regular_statement()), self.visit(ctx.substitutions(1))])
        while_statement = While(FileLocation(self.file_name,
                                             ctx.start, ctx.stop), self.visit(ctx.expression()), while_body)
        return Block(FileLocation(self.file_name, ctx.start, ctx.stop), [init, while_statement])

    # while_statement: WHILE LP expression RP regular_statement;
    def visitWhile_statement(self, ctx: CircomParser.While_statementContext):
        return While(FileLocation(self.file_name, ctx.start, ctx.stop), self.visit(ctx.expression()), self.visit(ctx.regular_statement()))

    # equal_constraint_statement: expression EQ_CONSTRAINT expression SEMICOLON;
    def visitEqual_constraint_statement(self, ctx: CircomParser.Equal_constraint_statementContext):
        return ConstraintEquality(FileLocation(self.file_name, ctx.start, ctx.stop), self.visit(ctx.expression(0)), self.visit(ctx.expression(1)))

    # return_statement: RETURN expression SEMICOLON;
    def visitReturn_statement(self, ctx: CircomParser.Return_statementContext):
        return Return(FileLocation(self.file_name, ctx.start, ctx.stop), self.visit(ctx.expression()))

    # assert_statement: ASSERT LP expression RP SEMICOLON;
    def visitAssert_statement(self, ctx: CircomParser.Assert_statementContext):
        return Assert(FileLocation(self.file_name, ctx.start, ctx.stop), self.visit(ctx.expression()))

    # log_statement
    #     : LOG LP log_list RP SEMICOLON
    #     | LOG LP RP SEMICOLON
    #     ;
    def visitLog_statement(self, ctx: CircomParser.Log_statementContext):
        return LogCall(FileLocation(self.file_name, ctx.start, ctx.stop), self.visit(ctx.log_list()) if ctx.log_list() else [])

    # declaration
    #     : var_decl
    #     | signal_decl
    #     | component_decl
    #     ;
    def visitDeclaration(self, ctx: CircomParser.DeclarationContext):
        return self.visit(ctx.getChild(0))

    # identifier_list: IDENTIFIER COMMA identifier_list | IDENTIFIER;
    def visitIdentifier_list(self, ctx: CircomParser.Identifier_listContext):
        return [ctx.IDENTIFIER().getText()] + self.visit(ctx.identifier_list()) if ctx.identifier_list() else [ctx.IDENTIFIER().getText()]

    # tag_list: LC identifier_list RC;
    def visitTag_list(self, ctx: CircomParser.Tag_listContext):
        return self.visit(ctx.identifier_list())

    # tuple_initialization: assign_opcode expression;
    def visitTuple_initialization(self, ctx: CircomParser.Tuple_initializationContext):
        return (self.visit(ctx.assign_opcode()), self.visit(ctx.expression()))

    # simple_symbol: IDENTIFIER array_access_list;
    def visitSimple_symbol(self, ctx: CircomParser.Simple_symbolContext):
        return (ctx.IDENTIFIER().getText(), self.visit(ctx.array_access_list()), None)

    # simple_symbol_list: simple_symbol COMMA simple_symbol_list | simple_symbol;
    def visitSimple_symbol_list(self, ctx: CircomParser.Simple_symbol_listContext):
        return [self.visit(ctx.simple_symbol())] + self.visit(ctx.simple_symbol_list()) if ctx.simple_symbol_list() else [self.visit(ctx.simple_symbol())]

    # complex_symbol: IDENTIFIER array_access_list ASSIGNMENT expression;
    def visitComplex_symbol(self, ctx: CircomParser.Complex_symbolContext):
        return (ctx.IDENTIFIER().getText(), self.visit(ctx.array_access_list()), self.visit(ctx.expression()))

    # some_symbol: simple_symbol | complex_symbol;
    def visitSome_symbol(self, ctx: CircomParser.Some_symbolContext):
        return self.visit(ctx.getChild(0))

    # some_symbol_list: some_symbol COMMA some_symbol_list | some_symbol;
    def visitSome_symbol_list(self, ctx: CircomParser.Some_symbol_listContext):
        return [self.visit(ctx.some_symbol())] + self.visit(ctx.some_symbol_list()) if ctx.some_symbol_list() else [self.visit(ctx.some_symbol())]

    # var_decl
    #     : VAR some_symbol_list
    #     | VAR LP simple_symbol_list RP
    #     | VAR LP simple_symbol_list RP tuple_initialization
    #     ;
    def visitVar_decl(self, ctx: CircomParser.Var_declContext):
        initializations = []
        if ctx.some_symbol_list():
            symbols = self.visit(ctx.some_symbol_list())
            for symbol in symbols:
                name, dimensions, init = symbol
                initializations.append(Declaration(FileLocation(self.file_name,
                                                                ctx.start, ctx.stop), Var(), name, dimensions, True))
                if init:
                    initializations.append(Substitution(FileLocation(self.file_name,
                                                                     ctx.start, ctx.stop), name, [], "=", init))
        else:
            values = []
            symbols = self.visit(ctx.simple_symbol_list())
            for symbol in symbols:
                name, dimensions, init = symbol
                initializations.append(Declaration(FileLocation(self.file_name,
                                                                ctx.start, ctx.stop), Var(), name, dimensions, True))
                values.append(Variable(FileLocation(self.file_name,
                                                    ctx.start, ctx.stop), name, []))
            if ctx.tuple_initialization():
                op, expr = self.visit(ctx.tuple_initialization())
                initializations.append(MultiSubstitution(FileLocation(self.file_name, ctx.start, ctx.stop), TupleExpr(
                    FileLocation(self.file_name, ctx.start, ctx.stop), values), op, expr))
        return InitializationBlock(FileLocation(self.file_name, ctx.start, ctx.stop), Var(), initializations)

    # signal_decl
    #     : signal_header signal_symbol_list
    #     | signal_header LP simple_symbol_list RP
    #     | signal_header LP simple_symbol_list RP tuple_initialization
    #     ;
    def visitSignal_decl(self, ctx: CircomParser.Signal_declContext):
        xtype = self.visit(ctx.signal_header())
        initializations = []
        if ctx.signal_symbol_list():
            symbols = self.visit(ctx.signal_symbol_list())
            for symbol in symbols:
                name, dimensions, init, sym_op = symbol
                initializations.append(Declaration(FileLocation(self.file_name,
                                                                ctx.start, ctx.stop), xtype, name, dimensions, True))
                if init:
                    initializations.append(Substitution(FileLocation(self.file_name,
                                                                     ctx.start, ctx.stop), name, [], sym_op, init))
        else:
            values = []
            symbols = self.visit(ctx.simple_symbol_list())
            for symbol in symbols:
                name, dimensions, init, sym_op = symbol
                initializations.append(Declaration(FileLocation(self.file_name,
                                                                ctx.start, ctx.stop), xtype, name, dimensions, True))
                values.append(Variable(FileLocation(self.file_name,
                                                    ctx.start, ctx.stop), name, []))
            if ctx.tuple_initialization():
                op, expr = self.visit(ctx.tuple_initialization())
                initializations.append(MultiSubstitution(FileLocation(self.file_name, ctx.start, ctx.stop), TupleExpr(
                    FileLocation(self.file_name, ctx.start, ctx.stop), values), op, expr))
        return InitializationBlock(FileLocation(self.file_name, ctx.start, ctx.stop), xtype, initializations)

    # signal_header
    #     : SIGNAL
    #     | SIGNAL SIGNAL_TYPE
    #     | SIGNAL tag_list
    #     | SIGNAL SIGNAL_TYPE tag_list
    #     ;
    def visitSignal_header(self, ctx: CircomParser.Signal_headerContext):
        return Signal(FileLocation(self.file_name, ctx.start, ctx.stop), ctx.SIGNAL_TYPE().getText() if ctx.SIGNAL_TYPE() else "intermediate", self.visit(ctx.tag_list()) if ctx.tag_list() else [])

    # signal_symbol
    #     : simple_symbol
    #     | IDENTIFIER array_access_list LEFT_CONSTRAINT expression
    #     | IDENTIFIER array_access_list LEFT_ASSIGNMENT expression
    #     ;
    def visitSignal_symbol(self, ctx: CircomParser.Signal_symbolContext):
        if ctx.simple_symbol():
            name, dimensions, init = self.visit(ctx.simple_symbol())
            op = ""
        else:
            name, dimensions, init = ctx.IDENTIFIER().getText(), self.visit(
                ctx.array_access_list()), self.visit(ctx.expression())
            op = ctx.LEFT_ASSIGNMENT().getText() if ctx.LEFT_ASSIGNMENT(
            ) else ctx.LEFT_CONSTRAINT().getText()
        return (name, dimensions, init, op)

    # signal_symbol_list: signal_symbol COMMA signal_symbol_list | signal_symbol;
    def visitSignal_symbol_list(self, ctx: CircomParser.Signal_symbol_listContext):
        return [self.visit(ctx.signal_symbol())] + self.visit(ctx.signal_symbol_list()) if ctx.signal_symbol_list() else [self.visit(ctx.signal_symbol())]

    # component_decl
    #     : COMPONENT some_symbol_list
    #     | COMPONENT LP simple_symbol_list RP
    #     | COMPONENT LP simple_symbol_list RP tuple_initialization
    #     ;
    def visitComponent_decl(self, ctx: CircomParser.Component_declContext):
        initializations = []
        if ctx.some_symbol_list():
            symbols = self.visit(ctx.some_symbol_list())
            for symbol in symbols:
                name, dimensions, init = symbol
                initializations.append(Declaration(FileLocation(self.file_name,
                                                                ctx.start, ctx.stop), Component(), name, dimensions, True))
                if init:
                    initializations.append(Substitution(FileLocation(self.file_name,
                                                                     ctx.start, ctx.stop), name, [], "=", init))
        else:
            values = []
            symbols = self.visit(ctx.simple_symbol_list())
            for symbol in symbols:
                name, dimensions, init = symbol
                initializations.append(Declaration(FileLocation(self.file_name,
                                                                ctx.start, ctx.stop), Component(), name, dimensions, True))
                values.append(Variable(FileLocation(self.file_name,
                                                    ctx.start, ctx.stop), name, []))
            if ctx.tuple_initialization():
                op, expr = self.visit(ctx.tuple_initialization())
                initializations.append(MultiSubstitution(FileLocation(self.file_name, ctx.start, ctx.stop), TupleExpr(
                    FileLocation(self.file_name, ctx.start, ctx.stop), values), op, expr))
        return InitializationBlock(FileLocation(self.file_name, ctx.start, ctx.stop), Component(), initializations)

    # var_access: array_access | component_access;
    def visitVar_access(self, ctx: CircomParser.Var_accessContext):
        return ArrayAccess(FileLocation(self.file_name, ctx.start, ctx.stop), self.visit(ctx.array_access())) if ctx.array_access() else ComponentAccess(FileLocation(self.file_name, ctx.start, ctx.stop), self.visit(ctx.component_access()))

    # var_access_list: var_access var_access_list | ;
    def visitVar_access_list(self, ctx: CircomParser.Var_access_listContext):
        return [self.visit(ctx.var_access())] + self.visit(ctx.var_access_list()) if ctx.getChildCount() == 2 else []

    # array_access: LB expression RB;
    def visitArray_access(self, ctx: CircomParser.Array_accessContext):
        return self.visit(ctx.expression())

    # array_access_list: array_access array_access_list | ;
    def visitArray_access_list(self, ctx: CircomParser.Array_access_listContext):
        return [self.visit(ctx.array_access())] + self.visit(ctx.array_access_list()) if ctx.array_access() else []

    # component_access: DOT IDENTIFIER;
    def visitComponent_access(self, ctx: CircomParser.Component_accessContext):
        return ctx.IDENTIFIER().getText()

    # variable: IDENTIFIER var_access_list;
    def visitVariable(self, ctx: CircomParser.VariableContext):
        return Variable(FileLocation(self.file_name, ctx.start, ctx.stop), ctx.IDENTIFIER().getText(), self.visit(ctx.var_access_list()))

    # expression
    #     : PARALLEL expression1
    #     | expression1
    #     ;
    def visitExpression(self, ctx: CircomParser.ExpressionContext):
        return ParallelOp(FileLocation(self.file_name, ctx.start, ctx.stop), self.visit(ctx.expression1())) if ctx.PARALLEL() else self.visit(ctx.expression1())

    # expression1
    #     : expression2 TERNARY_CONDITION expression2 TERNARY_ALTERNATIVE expression2
    #     | expression2
    #     ;
    def visitExpression1(self, ctx: CircomParser.Expression1Context):
        return InlineSwitchOp(FileLocation(self.file_name, ctx.start, ctx.stop), self.visit(ctx.expression2(0)), self.visit(ctx.expression2(1)), self.visit(ctx.expression2(2))) if ctx.TERNARY_ALTERNATIVE() else self.visit(ctx.expression2(0))

    # expression2
    #     : expression2 OR expression3
    #     | expression3
    #     ;
    def visitExpression2(self, ctx: CircomParser.Expression2Context):
        return InfixOp(FileLocation(self.file_name, ctx.start, ctx.stop), self.visit(ctx.expression2()), ctx.OR().getText(), self.visit(ctx.expression3())) if ctx.OR() else self.visit(ctx.expression3())

    # expression3
    #     : expression3 AND expression4
    #     | expression4
    #     ;
    def visitExpression3(self, ctx: CircomParser.Expression3Context):
        return InfixOp(FileLocation(self.file_name, ctx.start, ctx.stop), self.visit(ctx.expression3()), ctx.AND().getText(), self.visit(ctx.expression4())) if ctx.AND() else self.visit(ctx.expression4())

    # expression4
    #     : expression4 compareOpcode expression5
    #     | expression5
    #     ;
    def visitExpression4(self, ctx: CircomParser.Expression4Context):
        return InfixOp(FileLocation(self.file_name, ctx.start, ctx.stop), self.visit(ctx.expression4()), self.visit(ctx.compareOpcode()), self.visit(ctx.expression5())) if ctx.compareOpcode() else self.visit(ctx.expression5())

    # expression5
    #     : expression5 BOR expression6
    #     | expression6
    #     ;
    def visitExpression5(self, ctx: CircomParser.Expression5Context):
        return InfixOp(FileLocation(self.file_name, ctx.start, ctx.stop), self.visit(ctx.expression5()), ctx.BOR().getText(), self.visit(ctx.expression6())) if ctx.BOR() else self.visit(ctx.expression6())

    # expression6
    #     : expression6 BXOR expression7
    #     | expression7
    #     ;
    def visitExpression6(self, ctx: CircomParser.Expression6Context):
        return InfixOp(FileLocation(self.file_name, ctx.start, ctx.stop), self.visit(ctx.expression6()), ctx.BXOR().getText(), self.visit(ctx.expression7())) if ctx.BXOR() else self.visit(ctx.expression7())

    # expression7
    #     : expression7 BAND expression8
    #     | expression8
    #     ;
    def visitExpression7(self, ctx: CircomParser.Expression7Context):
        return InfixOp(FileLocation(self.file_name, ctx.start, ctx.stop), self.visit(ctx.expression7()), ctx.BAND().getText(), self.visit(ctx.expression8())) if ctx.BAND() else self.visit(ctx.expression8())

    # expression8
    #     : expression8 shiftOpcode expression9
    #     | expression9
    #     ;
    def visitExpression8(self, ctx: CircomParser.Expression8Context):
        return InfixOp(FileLocation(self.file_name, ctx.start, ctx.stop), self.visit(ctx.expression8()), self.visit(ctx.shiftOpcode()), self.visit(ctx.expression9())) if ctx.shiftOpcode() else self.visit(ctx.expression9())

    # expression9
    #     : expression9 add_sub_opcode expression10
    #     | expression10
    #     ;
    def visitExpression9(self, ctx: CircomParser.Expression9Context):
        return InfixOp(FileLocation(self.file_name, ctx.start, ctx.stop), self.visit(ctx.expression9()), self.visit(ctx.add_sub_opcode()), self.visit(ctx.expression10())) if ctx.add_sub_opcode() else self.visit(ctx.expression10())

    # expression10
    #     : expression10 mul_div_opcode expression11
    #     | expression11
    #     ;
    def visitExpression10(self, ctx: CircomParser.Expression10Context):
        return InfixOp(FileLocation(self.file_name, ctx.start, ctx.stop), self.visit(ctx.expression10()), self.visit(ctx.mul_div_opcode()), self.visit(ctx.expression11())) if ctx.mul_div_opcode() else self.visit(ctx.expression11())

    # expression11
    #     : expression11 POW expression12
    #     | expression12
    #     ;
    def visitExpression11(self, ctx: CircomParser.Expression11Context):
        return InfixOp(FileLocation(self.file_name, ctx.start, ctx.stop), self.visit(ctx.expression11()), ctx.POW().getText(), self.visit(ctx.expression12())) if ctx.POW() else self.visit(ctx.expression12())

    # expression12
    #     : prefixOpcode expression12
    #     | expression13
    #     ;
    def visitExpression12(self, ctx: CircomParser.Expression12Context):
        return PrefixOp(FileLocation(self.file_name, ctx.start, ctx.stop), self.visit(ctx.prefixOpcode()), self.visit(ctx.expression12())) if ctx.prefixOpcode() else self.visit(ctx.expression13())

    # expression13
    #     : IDENTIFIER LP listable RP LP listableAnon RP
    #     | IDENTIFIER LP listable RP
    #     | LB listable_prime RB
    #     | LP twoElemsListable RP
    #     | expression14
    #     ;
    def visitExpression13(self, ctx: CircomParser.Expression13Context):
        if ctx.listableAnon():
            listableAnon = self.visit(ctx.listableAnon())
            signals = []
            names = []
            for anon in listableAnon:
                signals.append(anon[0])
                names.append(anon[1])
            return AnonymousComponentExpr(FileLocation(self.file_name, ctx.start, ctx.stop), ctx.IDENTIFIER().getText(), False, self.visit(ctx.listable()), signals, names)
        elif ctx.getChildCount() == 4:
            return Call(FileLocation(self.file_name, ctx.start, ctx.stop), ctx.IDENTIFIER().getText(), self.visit(ctx.listable()))
        elif ctx.getChildCount() == 3:
            return ArrayInLine(FileLocation(self.file_name, ctx.start, ctx.stop), self.visit(ctx.listable_prime())) if ctx.listable_prime() else TupleExpr(FileLocation(self.file_name, ctx.start, ctx.stop), self.visit(ctx.twoElemsListable()))
        else:
            return self.visit(ctx.expression14())

    # expression14
    #     : variable
    #     | UNDERSCORE
    #     | NUMBER
    #     | HEXNUMBER
    #     | LP expression RP
    #     ;
    def visitExpression14(self, ctx: CircomParser.Expression14Context):
        if ctx.variable():
            return self.visit(ctx.variable())
        elif ctx.UNDERSCORE():
            return Variable(FileLocation(self.file_name, ctx.start, ctx.stop), ctx.UNDERSCORE().getText(), [])
        elif ctx.NUMBER():
            return Number(FileLocation(self.file_name, ctx.start, ctx.stop), int(ctx.NUMBER().getText(), 10))
        elif ctx.HEXNUMBER():
            return Number(FileLocation(self.file_name, ctx.start, ctx.stop), int(ctx.HEXNUMBER().getText(), 16))
        else:
            return self.visit(ctx.expression())

    # twoElemsListable
    #     : twoElemsListable COMMA expression
    #     | expression COMMA expression
    #     ;
    def visitTwoElemsListable(self, ctx: CircomParser.TwoElemsListableContext):
        return self.visit(ctx.twoElemsListable()) + [self.visit(ctx.expression(0))] if ctx.twoElemsListable() else [self.visit(ctx.expression(0)), self.visit(ctx.expression(1))]

    # log_arguement: expression | STRING;
    def visitLog_arguement(self, ctx: CircomParser.Log_arguementContext):
        return LogExp(FileLocation(self.file_name, ctx.start, ctx.stop), self.visit(ctx.expression())) if ctx.expression() else LogStr(FileLocation(self.file_name, ctx.start, ctx.stop), ctx.STRING().getText())

    # log_list: log_arguement COMMA log_list | log_arguement;
    def visitLog_list(self, ctx: CircomParser.Log_listContext):
        return [self.visit(ctx.log_arguement())] + self.visit(ctx.log_list()) if ctx.log_list() else [self.visit(ctx.log_arguement())]

    # listable: listable_prime | ;
    def visitListable(self, ctx: CircomParser.ListableContext):
        return self.visit(ctx.listable_prime()) if ctx.listable_prime() else []

    # listable_prime: expression | expression COMMA listable_prime;
    def visitListable_prime(self, ctx: CircomParser.Listable_primeContext):
        return [self.visit(ctx.expression())] + self.visit(ctx.listable_prime()) if ctx.listable_prime() else [self.visit(ctx.expression())]

    # listableAnon: listableAnon_prime | ;
    def visitListableAnon(self, ctx: CircomParser.ListableAnonContext):
        return self.visit(ctx.listableAnon_prime()) if ctx.listableAnon_prime() else []

    # listableAnon_prime: listable_prime | listableWithInputNames;
    def visitListableAnon(self, ctx: CircomParser.ListableAnon_primeContext):
        return list(map(lambda x: (x, None), self.visit(ctx.listable_prime()))) if ctx.listable_prime() else self.visit(ctx.listableWithInputNames())

    # listableWithInputNames
    #     : listableWithInputNames COMMA IDENTIFIER assign_opcode expression
    #     | IDENTIFIER assign_opcode expression
    #     ;

    def visitListableWithInputNames(self, ctx: CircomParser.ListableWithInputNamesContext):
        return self.visit(ctx.listableWithInputNames()) + [(self.visit(ctx.expression()), (self.visit(ctx.assign_opcode()), ctx.IDENTIFIER().getText()))] if ctx.listableWithInputNames() else [(self.visit(ctx.expression()), (self.visit(ctx.assign_opcode()), ctx.IDENTIFIER().getText()))]

    # prefixOpcode: NOT | BNOT | SUB;
    def visitPrefixOpcode(self, ctx: CircomParser.PrefixOpcodeContext):
        return ctx.getChild(0).getText()

    # compareOpcode
    #     : EQ
    #     | NEQ
    #     | GT
    #     | LT
    #     | GE
    #     | LE
    #     ;
    def visitCompareOpcode(self, ctx: CircomParser.CompareOpcodeContext):
        return ctx.getChild(0).getText()

    # shiftOpcode: SHL | SHR;
    def visitShiftOpcode(self, ctx: CircomParser.ShiftOpcodeContext):
        return ctx.getChild(0).getText()

    # add_sub_opcode: ADD | SUB;
    def visitAdd_sub_opcode(self, ctx: CircomParser.Add_sub_opcodeContext):
        return ctx.getChild(0).getText()

    # mul_div_opcode: MUL | DIV | QUO | MOD;
    def visitMul_div_opcode(self, ctx: CircomParser.Mul_div_opcodeContext):
        return ctx.getChild(0).getText()

    # assign_opcode: ASSIGNMENT | LEFT_ASSIGNMENT | LEFT_CONSTRAINT;
    def visitAssign_opcode(self, ctx: CircomParser.Assign_opcodeContext):
        return ctx.getChild(0).getText()



from AST import *
from Visitor import *
from Errors import *
from Report import Report, ReportType
from enum import Enum, auto


class Symbol:
    def __init__(self, name, mtype, xtype, ast=None, value=None):
        self.name = name
        self.mtype = mtype  # PrimeField, ArrayCircom, TemplateCircom, FunctionCircom
        self.xtype = xtype  # VarCircom, SignalCircom, ComponentCircom, AnonymousComponentCircom
        self.ast = ast
        self.value = value


class Type:
    pass


class SignalType(Enum):
    INPUT = auto()
    OUTPUT = auto()
    INTERMEDIATE = auto()


class PrimeField(Type):
    pass


class SignalCircom(Type):
    def __init__(self, signal_type):
        self.signal_type = signal_type


class VarCircom(Type):
    pass


class ComponentCircom(Type):
    pass


class AnonymousComponentCircom(Type):
    pass


class ArrayCircom(Type):
    def __init__(self, eleType, dims):
        self.dims = dims
        self.eleType = eleType


class TemplateCircom(Type):
    def __init__(self, name, params, signals={}, signals_in=[], signals_out=[], args=[], signals_ast={}):
        self.name = name
        self.params = params
        self.signals = signals
        self.signals_in = signals_in
        self.signals_out = signals_out
        self.args = args
        self.signals_ast = signals_ast


class FunctionCircom(Type):
    def __init__(self, name, params, return_type, body, args):
        self.name = name
        self.params = params
        self.return_type = return_type
        self.body = body
        self.args = args


def is_same_type(type1, type2):
    if type(type1) != type(type2):
        return False
    if isinstance(type1, ArrayCircom) and isinstance(type2, ArrayCircom):
        return type1.dims == type2.dims and is_same_type(type1.eleType, type2.eleType)
    return True


class TypeCheck(BaseVisitor):
    def __init__(self, ast):
        self.ast = ast
        self.global_env = [{}]
        self.list_function = {}
        self.in_function = False
        self.no_block = False
        self.return_func = None
        self.list_template = {}
        self.in_template = False
        self.template_signals = None
        self.signals_ast = None
        self.signal_in = None
        self.signal_out = None
        self.is_multi_sub = False
        self.count_visited = 0

    def infer_type(self, var, param, typ=PrimeField()):
        symbol = None
        for env in param:
            if var.name in env:
                symbol = env[var.name]
                break
        if isinstance(typ, PrimeField):
            if len(var.access) == 0:
                symbol.mtype = typ
            else:
                symbol.mtype = ArrayCircom(typ, len(var.access))
        elif isinstance(typ, ArrayCircom):
            if len(var.access) == 0:
                symbol.mtype = typ
            else:
                symbol.mtype = ArrayCircom(
                    typ.eleType, len(var.access) + typ.dims)
        self.visit(var, param)

    def check(self):
        self.visit(self.ast, self.global_env)

    def visitFileLocation(self, ast: FileLocation, param):
        return None

    def visitMainComponent(self, ast: MainComponent, param):
        template_type = self.visit(ast.expr, param)
        if not isinstance(template_type, TemplateCircom):
            raise Report(ReportType.ERROR, ast.locate,
                         f"Main component must be a template. Found {type(template_type).__name__} instead")
        for name in ast.publics:
            if name not in template_type.signals_in:
                raise Report(ReportType.ERROR, ast.locate,
                             f"Signal '{name}' not declared in template '{template_type.name}'")
        param[0]["main"] = Symbol(
            "main", template_type, ComponentCircom(), ast)

    def visitInclude(self, ast: Include, param):
        return None

    def visitTemplate(self, ast: Template, param):
        if self.count_visited == 0:
            if ast.name_field in self.list_template:
                raise Report(ReportType.ERROR, ast.locate,
                             f"Template '{ast.name_field}' already declared")
            env = [{}] + param
            for arg in ast.args:
                if arg in env[0]:
                    raise Report(ReportType.ERROR, ast.locate,
                                 f"Argument '{arg}' already declared")
                env[0][arg] = Symbol(arg, None, VarCircom())
            self.in_template = True
            self.template_signals = {}
            self.signals_ast = {}
            self.signal_in = []
            self.signal_out = []
            self.visit(ast.body, env)
            if self.return_func is not None:
                raise Report(ReportType.ERROR, ast.locate,
                             "Template can not have a return statement")
            self.in_template = False
            arg_list = []
            for arg in ast.args:
                arg_list.append(env[0][arg].mtype)
            self.list_template[ast.name_field] = param[0][ast.name_field] = Symbol(ast.name_field, TemplateCircom(
                ast.name_field, arg_list, self.template_signals, self.signal_in, self.signal_out, ast.args, self.signals_ast), None, ast)
            self.template_signals = None
            self.signals_ast = None
            self.signal_in = None
            self.signal_out = None
        else:
            env = [{}] + param
            for arg in ast.args:
                env[0][arg] = Symbol(arg, None, VarCircom())
            self.in_template = True
            self.visit(ast.body, env)
            arg_list = []
            for arg in ast.args:
                arg_list.append(env[0][arg].mtype)
            self.list_template[ast.name_field].mtype.params = arg_list
            if self.return_func is not None:
                raise Report(ReportType.ERROR, ast.locate,
                             "Template can not have a return statement")
            self.in_template = False

    def visitFunction(self, ast: Function, param):
        if self.count_visited == 0:
            if ast.name_field in self.list_function:
                raise Report(ReportType.ERROR, ast.locate,
                             f"Function '{ast.name_field}' already declared")
            env = [{}] + param
            for arg in ast.args:
                if arg in env[0]:
                    raise Report(ReportType.ERROR, ast.locate,
                                 f"Argument '{arg}' already declared")
                env[0][arg] = Symbol(arg, None, VarCircom())
            self.in_function = True
            self.no_block = True
            self.visit(ast.body, env)
            self.in_function = False
            if self.return_func is None:
                raise Report(ReportType.ERROR, ast.locate,
                             "Unable to infer the type of this function")
            arg_list = []
            for arg in ast.args:
                arg_list.append(env[0][arg].mtype)
            self.list_function[ast.name_field] = param[0][ast.name_field] = Symbol(ast.name_field, FunctionCircom(
                ast.name_field, arg_list, self.return_func, ast.body, ast.args), None, ast)
            self.return_func = None
        else:
            self.in_function = True
            self.no_block = True
            env = [{}] + param
            for arg in ast.args:
                env[0][arg] = Symbol(arg, None, VarCircom())
            self.visit(ast.body, env)
            if self.return_func is None:
                raise Report(ReportType.ERROR, ast.locate,
                             "Unable to infer the type of this function")
            self.return_func = None
            self.in_function = False

    def visitProgram(self, ast: Program, param):
        for _ in range(2):
            for definition in ast.definitions:
                self.visit(definition, param)
            self.count_visited += 1
        if ast.main_component is None:
            raise Report(ReportType.ERROR, ast.locate,
                         "Main component not declared")
        self.visit(ast.main_component, param)

    def visitIfThenElse(self, ast: IfThenElse, param):
        if self.count_visited == 0:
            self.visit(ast.if_case, param)
            if ast.else_case:
                self.visit(ast.else_case, param)
                return
        cond_type = self.visit(ast.cond, param)
        if isinstance(cond_type, TemplateCircom):
            raise Report(ReportType.ERROR, ast.cond.locate,
                         "Must be a single arithmetic expression. Found component")
        elif isinstance(cond_type, ArrayCircom):
            raise Report(ReportType.ERROR, ast.cond.locate,
                         "Must be a single arithmetic expression. Found array")
        self.visit(ast.if_case, param)
        if ast.else_case:
            self.visit(ast.else_case, param)

    def visitWhile(self, ast: While, param):
        if self.count_visited == 0:
            self.visit(ast.stmt, param)
            return
        cond_type = self.visit(ast.cond, param)
        if isinstance(cond_type, TemplateCircom):
            raise Report(ReportType.ERROR, ast.cond.locate,
                         "Must be a single arithmetic expression. Found component")
        elif isinstance(cond_type, ArrayCircom):
            raise Report(ReportType.ERROR, ast.cond.locate,
                         "Must be a single arithmetic expression. Found array")
        self.visit(ast.stmt, param)

    def visitReturn(self, ast: Return, param):
        if self.in_function:
            value_type = self.visit(ast.value, param)
            if isinstance(value_type, TemplateCircom):
                raise Report(ReportType.ERROR, ast.value.locate,
                             "Must be a single arithmetic expression. Found component")
            if self.return_func is None:
                self.return_func = value_type
            if not is_same_type(value_type, self.return_func):
                raise Report(ReportType.ERROR, ast.value.locate,
                             "Return type is not compatible with the function return type")
        else:
            raise Report(ReportType.ERROR, ast.locate,
                         "Return statement outside of a function")

    def visitInitializationBlock(self, ast: InitializationBlock, param):
        for init in ast.initializations:
            if self.count_visited == 0:
                if isinstance(init, Declaration):
                    self.visit(init, param)
            else:
                self.visit(init, param)

    def visitDeclaration(self, ast: Declaration, param):
        if ast.name in param[0]:
            raise Report(ReportType.ERROR, ast.locate,
                         f"Variable '{ast.name}' already declared")
        for expr in ast.dimensions:
            expr_type = self.visit(expr, param)
            if isinstance(expr_type, TemplateCircom):
                raise Report(ReportType.ERROR, expr.locate,
                             "Array indexes and lengths must be single arithmetic expressions. Found component instead of expression.")
            elif isinstance(expr_type, ArrayCircom):
                raise Report(ReportType.ERROR, expr.locate,
                             "Array indexes and lengths must be single arithmetic expressions. Found array instead of expression.")
        xtype = self.visit(ast.xtype, param)
        if isinstance(xtype, SignalCircom):
            if self.count_visited == 0:
                if ast.name in self.template_signals:
                    raise Report(ReportType.ERROR, ast.locate,
                                 f"Signal '{ast.name}' already declared")
                if xtype.signal_type == SignalType.INPUT:
                    self.signal_in.append(ast.name)
                elif xtype.signal_type == SignalType.OUTPUT:
                    self.signal_out.append(ast.name)
            if len(ast.dimensions) > 0:
                mtype = ArrayCircom(PrimeField(), len(ast.dimensions))
            else:
                mtype = PrimeField()
            if self.count_visited == 0:
                self.template_signals[ast.name] = mtype
                self.signals_ast[ast.name] = ast
            param[0][ast.name] = Symbol(ast.name, mtype, xtype, ast)
        elif isinstance(xtype, VarCircom):
            if len(ast.dimensions) > 0:
                mtype = ArrayCircom(PrimeField(), len(ast.dimensions))
            else:
                mtype = PrimeField()
            param[0][ast.name] = Symbol(ast.name, mtype, xtype, ast)
        elif isinstance(xtype, ComponentCircom):
            template_type = TemplateCircom("", None)
            if len(ast.dimensions) > 0:
                mtype = ArrayCircom(template_type, len(ast.dimensions))
            else:
                mtype = template_type
            param[0][ast.name] = Symbol(ast.name, mtype, xtype, ast)

    def visitSubstitution(self, ast: Substitution, param):
        if self.count_visited == 0:
            return
        symbol = None
        for env in param:
            if ast.var in env:
                symbol = env[ast.var]
                break
        if self.is_multi_sub:
            self.is_multi_sub = False
            rhe_type = ast.rhe
        else:
            rhe_type = self.visit(ast.rhe, param)
        if ast.var == "_":
            if isinstance(rhe_type, TemplateCircom):
                raise Report(ReportType.ERROR, ast.rhe.locate,
                             "Must be a single arithmetic expression. Found component")
        else:
            lhe_type = self.visit(
                Variable(ast.locate, ast.var, ast.access), param)
            symbol = None
            for env in param:
                if ast.var in env:
                    symbol = env[ast.var]
                    break
            if isinstance(symbol.xtype, SignalCircom):
                if symbol.xtype.signal_type == SignalType.INPUT:
                    raise Report(ReportType.ERROR, ast.locate,
                                 "Cannot assign to an input signal")
                if ast.op == "=":
                    raise Report(ReportType.ERROR, ast.locate,
                                 "Cannot assign to a signal")
                if rhe_type is None:
                    self.infer_type(ast.rhe, param, lhe_type)
                    rhe_type = lhe_type
                if isinstance(rhe_type, TemplateCircom):
                    raise Report(ReportType.ERROR, ast.rhe.locate,
                                 "Must be a single arithmetic expression. Found component")
                if not is_same_type(lhe_type, rhe_type):
                    raise Report(ReportType.ERROR, ast.rhe.locate,
                                 "Types of the two sides of the equality are not compatible")
            elif isinstance(symbol.xtype, VarCircom):
                if ast.op != "=":
                    raise Report(ReportType.ERROR, ast.locate,
                                 "Cannot use operator on a variable")
                if rhe_type is None:
                    self.infer_type(ast.rhe, param, lhe_type)
                    rhe_type = lhe_type
                if isinstance(rhe_type, TemplateCircom):
                    raise Report(ReportType.ERROR, ast.rhe.locate,
                                 "Must be a single arithmetic expression. Found component")
                if not is_same_type(lhe_type, rhe_type):
                    raise Report(ReportType.ERROR, ast.rhe.locate,
                                 "Types of the two sides of the equality are not compatible")
            elif isinstance(symbol.xtype, ComponentCircom):
                if isinstance(lhe_type, TemplateCircom):
                    if not isinstance(rhe_type, TemplateCircom):
                        raise Report(ReportType.ERROR, ast.rhe.locate,
                                     "Assignee and assigned types do not match.\n Expected template but found expression.")
                    if ast.op != "=":
                        raise Report(ReportType.ERROR, ast.locate,
                                     "Cannot use operator on a component")
                    if lhe_type.name == "":
                        lhe_type.name = rhe_type.name
                        lhe_type.params = rhe_type.params
                        lhe_type.signals = rhe_type.signals
                        lhe_type.signals_in = rhe_type.signals_in
                        lhe_type.signals_out = rhe_type.signals_out
                        lhe_type.args = rhe_type.args
                    else:
                        if lhe_type.name != rhe_type.name:
                            raise Report(ReportType.ERROR, ast.locate,
                                         "Assignee and assigned types do not match.")
                elif not is_same_type(lhe_type, rhe_type):
                    if rhe_type is None:
                        self.infer_type(ast.rhe, param, lhe_type)
                        return
                    raise Report(ReportType.ERROR, ast.rhe.locate,
                                 "Assignee and assigned types do not match.")

    def visitMultiSubstitution(self, ast: MultiSubstitution, param):
        if not isinstance(ast.lhe, TupleExpr):
            raise Report(ReportType.ERROR, ast.lhe.locate,
                         "Left hand side of the substitution must be a tuple")
        if not isinstance(ast.rhe, AnonymousComponentExpr):
            raise Report(ReportType.ERROR, ast.rhe.locate,
                         "Right hand side of the substitution must be a component")
        lhe_type = self.visit(ast.lhe, param)
        rhe_type = self.visit(ast.rhe, param)
        if len(lhe_type) != len(rhe_type):
            raise Report(ReportType.ERROR, ast.rhe.locate,
                         "Number of elements in the left hand side and right hand side must be the same")
        for i in range(len(lhe_type)):
            if not isinstance(lhe_type[i], Variable):
                raise Report(ReportType.ERROR, ast.lhe.locate, "Not variable")
            if lhe_type.name == "_":
                continue
            self.is_multi_sub = True
            self.visit(Substitution(
                ast.lhe[i].locate, lhe_type[i].name, lhe_type[i].access, ast.op, rhe_type[i]), param)

    def visitConstraintEquality(self, ast: ConstraintEquality, param):
        lhe_type = self.visit(ast.lhe, param)
        rhe_type = self.visit(ast.rhe, param)
        if isinstance(lhe_type, TemplateCircom):
            raise Report(ReportType.ERROR, ast.lhe.locate,
                         "Must be a single arithmetic expression. Found component")
        if isinstance(rhe_type, TemplateCircom):
            raise Report(ReportType.ERROR, ast.rhe.locate,
                         "Must be a single arithmetic expression. Found component")
        if not is_same_type(lhe_type, rhe_type):
            raise Report(ReportType.ERROR, ast.rhe.locate,
                         "Types of the two sides of the equality are not compatible")

    def visitLogCall(self, ast: LogCall, param):
        for arg in ast.args:
            if isinstance(arg, LogExp):
                arg_type = self.visit(arg, param)
                if isinstance(arg_type, TemplateCircom):
                    raise Report(ReportType.ERROR, arg.locate,
                                 "Must be a single arithmetic expression. Found component")
                elif isinstance(arg_type, ArrayCircom):
                    raise Report(ReportType.ERROR, arg.locate,
                                 "Must be a single arithmetic expression. Found array")

    def visitBlock(self, ast: Block, param):
        if self.in_template or self.no_block:
            env = param
            self.in_template = False
            self.no_block = False
        else:
            env = [{}] + param
        for stmt in ast.stmts:
            if self.count_visited == 0:
                if isinstance(stmt, InitializationBlock) or isinstance(stmt, Return) or isinstance(stmt, Block) or isinstance(stmt, IfThenElse) or isinstance(stmt, While):
                    self.visit(stmt, env)
            else:
                self.visit(stmt, env)

    def visitAssert(self, ast: Assert, param):
        arg_type = self.visit(ast.arg, param)
        if isinstance(arg_type, TemplateCircom):
            raise Report(ReportType.ERROR, ast.locate,
                         "Must be a single arithmetic expression. Found component")
        elif isinstance(arg_type, ArrayCircom):
            raise Report(ReportType.ERROR, ast.locate,
                         "Must be a single arithmetic expression. Found array")

    def visitVar(self, ast: Var, param):
        return VarCircom()

    def visitSignal(self, ast: Signal, param):
        if ast.signal_type == "input":
            signal_type = SignalType.INPUT
        elif ast.signal_type == "output":
            signal_type = SignalType.OUTPUT
        elif ast.signal_type == "intermediate":
            signal_type = SignalType.INTERMEDIATE
        return SignalCircom(signal_type)

    def visitComponent(self, ast: Component, param):
        return ComponentCircom()

    def visitAnonymousComponent(self, ast: AnonymousComponent, param):
        return AnonymousComponentCircom()

    def visitInfixOp(self, ast: InfixOp, param):
        lhe_type = self.visit(ast.lhe, param)
        if lhe_type is None:
            self.infer_type(ast.lhe, param)
            lhe_type = PrimeField()
        rhe_type = self.visit(ast.rhe, param)
        if rhe_type is None:
            self.infer_type(ast.rhe, param)
            rhe_type = PrimeField()
        if isinstance(lhe_type, TemplateCircom) or isinstance(lhe_type, ArrayCircom):
            raise Report(ReportType.ERROR, ast.lhe.locate,
                         "Type not allowed by the operator")
        if isinstance(rhe_type, TemplateCircom) or isinstance(rhe_type, ArrayCircom):
            raise Report(ReportType.ERROR, ast.rhe.locate,
                         "Type not allowed by the operator")
        return PrimeField()

    def visitPrefixOp(self, ast: PrefixOp, param):
        rhe_type = self.visit(ast.rhe, param)
        if rhe_type is None:
            self.infer_type(ast.rhe, param)
            rhe_type = PrimeField()
        if isinstance(rhe_type, TemplateCircom) or isinstance(rhe_type, ArrayCircom):
            raise Report(ReportType.ERROR, ast.rhe.locate,
                         "Type not allowed by the operator")
        return PrimeField()

    def visitInlineSwitchOp(self, ast: InlineSwitchOp, param):
        cond_type = self.visit(ast.cond, param)
        if isinstance(cond_type, TemplateCircom):
            raise Report(ReportType.ERROR, ast.cond.locate,
                         "Must be a single arithmetic expression. Found component")
        elif isinstance(cond_type, ArrayCircom):
            raise Report(ReportType.ERROR, ast.cond.locate,
                         "Must be a single arithmetic expression. Found array")
        if_true_type = self.visit(ast.if_true, param)
        if_false_type = self.visit(ast.if_false, param)
        if not is_same_type(if_true_type, if_false_type):
            raise Report(ReportType.ERROR, ast.if_false.locate,
                         "Inline switch operator branches types are non compatible")
        return if_true_type

    def visitParrallelOp(self, ast: ParallelOp, param):
        rhe_type = self.visit(ast.rhe, param)
        if not isinstance(rhe_type, TemplateCircom):
            raise Report(ReportType.ERROR, ast.rhe.locate,
                         "Type not allowed by the operator parallel (parallel operator can only be applied to templates)")
        return rhe_type

    def visitVariable(self, ast: Variable, param):
        symbol = None
        for env in param:
            if ast.name in env:
                symbol = env[ast.name]
                break
        if symbol is None:
            raise Report(ReportType.ERROR, ast.locate,
                         f"Variable '{ast.name}' not declared")
        typ = symbol.mtype
        if typ is None:
            return None
        if isinstance(typ, ArrayCircom):
            var_type = ArrayCircom(typ.eleType, typ.dims)
        else:
            var_type = typ
        for i in range(len(ast.access)):
            if isinstance(var_type, PrimeField):
                raise Report(ReportType.ERROR, ast.locate,
                             f"Variable '{ast.name}' is not an array or component")
            access_type = self.visit(ast.access[i], param)
            if isinstance(access_type, PrimeField):
                if not isinstance(var_type, ArrayCircom):
                    raise Report(ReportType.ERROR, ast.locate,
                                 f"Variable '{ast.name}' is not an array")
                if var_type.dims == 1:
                    var_type = var_type.eleType
                else:
                    var_type.dims -= 1
            else:
                if not isinstance(var_type, TemplateCircom):
                    raise Report(ReportType.ERROR, ast.locate,
                                 f"Variable '{ast.name}' is not a component")
                # print(ast.locate)
                signal_type = var_type.signals[access_type]
                if isinstance(signal_type, ArrayCircom):
                    var_type = ArrayCircom(
                        signal_type.eleType, signal_type.dims)
                else:
                    var_type = signal_type
        return var_type

    def visitNumber(self, ast: Number, param):
        return PrimeField()

    def visitCall(self, ast: Call, param):
        if ast.id in self.list_function:
            func_type = self.list_function[ast.id].mtype
        elif ast.id in self.list_template:
            func_type = self.list_template[ast.id].mtype
        else:
            raise Report(ReportType.ERROR, ast.locate,
                         f"Function or Template '{ast.id}' not declared")
        if not isinstance(func_type, (FunctionCircom, TemplateCircom)):
            raise Report(ReportType.ERROR, ast.locate,
                         f"'{ast.id}' is not a function or template")
        if len(ast.args) != len(func_type.params):
            raise Report(ReportType.ERROR, ast.locate,
                         f"Function '{ast.id}' has {len(func_type.params)} arguments, but {len(ast.args)} were provided")
        for arg in ast.args:
            self.visit(arg, param)
        if isinstance(func_type, FunctionCircom):
            return func_type.return_type
        elif isinstance(func_type, TemplateCircom):
            return func_type

    def visitAnonymousComponentExpr(self, ast: AnonymousComponentExpr, param):
        template_type = None
        for env in param:
            if ast.id in env:
                template_type = env[ast.id].mtype
                break
        if template_type is None or not isinstance(template_type, TemplateCircom):
            raise Report(ReportType.ERROR, ast.locate,
                         f"Template '{ast.id}' not declared")
        if len(ast.params) != len(template_type.params):
            raise Report(ReportType.ERROR, ast.locate,
                         "Don't have same size of params")
        for i in range(len(ast.params)):
            param_type = self.visit(ast.params[i], param)
            if not is_same_type(param_type, template_type.params[i]):
                raise Report(ReportType.ERROR, par.locate,
                             "Type Mismatch in parameter")
        if ast.names and ast.names[0]:
            if len(ast.names) != len(template_type.signals_in):
                raise Report(ReportType.ERROR, ast.locate,
                             f"Template '{ast.id}' has {len(template_type.signals_in)} input signals, but {len(ast.names)} were provided")
            for i in range(len(ast.names)):
                op, name = ast.names[i]
                if name not in template_type.signals:
                    raise Report(ReportType.ERROR, ast.locate,
                                 f"Signal '{name}' not declared in template '{ast.id}'")
                if name not in template_type.signals_in:
                    raise Report(ReportType.ERROR, ast.locate,
                                 f"Signal '{name}' is not an input signal")
                if not is_same_type(template_type.signals[name], self.visit(ast.signals[i], param)):
                    raise Report(ReportType.ERROR, ast.locate,
                                 f"Signal '{name}' type does not match the template '{ast.id}' input signal type")
        else:
            if len(ast.signals) != len(template_type.signals_in):
                raise Report(ReportType.ERROR, ast.locate,
                             f"Template '{ast.id}' has {len(template_type.signals_in)} input signals, but {len(ast.signals)} were provided")
            for i in range(len(ast.signals)):
                if not is_same_type(template_type.signals[template_type.signals_in[i]], self.visit(ast.signals[i], param)):
                    raise Report(ReportType.ERROR, ast.locate,
                                 f"Signal '{template_type.signals_in[i]}' type does not match the template '{ast.id}' input signal type")
        return_tuple = []
        for name in template_type.signals_out:
            return_tuple.append(template_type.signals[name])
        return return_tuple if len(return_tuple) > 1 else return_tuple[0]

    def visitArrayInLine(self, ast: ArrayInLine, param):
        type_list = []
        for value in ast.values:
            type_list.append(self.visit(value, param))
        first_type = type_list[0]
        if isinstance(first_type, TemplateCircom):
            raise Report(
                ReportType.ERROR, ast.values[0].locate, "Components can not be declared inside inline arrays")
        for i in range(1, len(type_list)):
            if isinstance(type_list[i], TemplateCircom):
                raise Report(
                    ReportType.ERROR, ast.values[i].locate, "Components can not be declared inside inline arrays")
            elif not is_same_type(first_type, type_list[i]):
                raise Report(ReportType.ERROR, ast.values[i].locate,
                             "All elements in the array must be of the same type")
        if isinstance(first_type, PrimeField):
            return ArrayCircom(first_type, 1)
        elif isinstance(first_type, ArrayCircom):
            return ArrayCircom(first_type.eleType, first_type.dims + 1)

    def visitTupleExpr(self, ast: TupleExpr, param):
        return ast.values

    def visitComponentAccess(self, ast: ComponentAccess, param):
        return ast.name

    def visitArrayAccess(self, ast: ArrayAccess, param):
        expr_type = self.visit(ast.expr, param)
        if isinstance(expr_type, TemplateCircom):
            raise Report(ReportType.ERROR, ast.expr.locate,
                         "Array indexes and lengths must be single arithmetic expressions. Found component instead of expression.")
        elif isinstance(expr_type, ArrayCircom):
            raise Report(ReportType.ERROR, ast.expr.locate,
                         "Array indexes and lengths must be single arithmetic expressions. Found array instead of expression.")
        return PrimeField()

    def visitLogStr(self, ast: LogStr, param):
        return None

    def visitLogExp(self, ast: LogExp, param):
        return self.visit(ast.expr, param)



from functools import wraps
import time
from antlr4 import *
import sys
import traceback
import os


def timeit(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        print(f"[Timeit]     Analysis completed in {(end - start):.2f} s")
        return result
    return wrapper


def generate_ast(filename):
    from CircomLexer import CircomLexer
    from CircomParser import CircomParser
    from ASTGeneration import ASTGeneration
    import Errors
    from Report import Report, ReportType
    from AST import FileLocation

    try:
        inputstream = FileStream(filename, encoding='utf-8')
        lexer = CircomLexer(inputstream)
        listener = Errors.NewErrorListener.INSTANCE
        tokens = CommonTokenStream(lexer)

        parser = CircomParser(tokens)
        parser.removeErrorListeners()
        parser.addErrorListener(listener)

        try:
            tree = parser.program()
        except Errors.SyntaxException as f:
            Report(ReportType.ERROR, FileLocation(
                filename, None, None), f.message, False).show()
            return None
        except Exception as e:
            Report(ReportType.ERROR, FileLocation(
                filename, None, None), str(e), False).show()
            return None
        path_env = [filename]
        asttree = ASTGeneration().generate_ast(filename, tree, path_env)
        return asttree
    except Exception as e:
        Report(ReportType.ERROR, FileLocation(filename, None, None),
               "Error generating AST: " + str(e), False).show()
        return None


def typecheck(ast):
    from StaticCheck import TypeCheck
    from Report import Report, ReportType
    from AST import FileLocation
    checked = TypeCheck(ast)
    try:
        checked.check()
    except Exception as e:
        # traceback.print_exc()
        print(e)
        return None, None, None
    return checked.global_env, checked.list_function, checked.list_template


def generate_cdg(ast, param, list_function, list_template):
    from CDGGeneration import CDGGeneration
    try:
        graphs = CDGGeneration(ast, param, list_function,
                               list_template).generateCDG()
        return graphs
    except Exception as e:
        traceback.print_exc()
        print(e)
        return None


def print_reports(graphs, reports):
    from Report import Report, ReportType
    from colorama import Fore
    to_print = {}
    total_reports = 0
    for graph in graphs.values():
        for report_list in reports[graph.name].values():
            for report in report_list:
                path = report.location.path
                line = report.location.start.line
                col = report.location.start.column
                path += f":{line}:{col}"
                if path not in to_print:
                    total_reports += 1
                    to_print[path] = Report(
                        report.type, report.location, report.message)
                else:
                    if report.message not in to_print[path].message:
                        total_reports += 1
                        to_print[path].message += "\n" + \
                            ' ' * 13 + report.message
    print("\n" + "=" * 50)
    for report in to_print.values():
        report.show()
    print(Fore.YELLOW + f"⚠ Total warnings: {total_reports}")


def report_to_file(graphs, reports, filename="report.json"):
    import json
    output = {}
    for graph in graphs.values():
        has_report = False
        for report_list in reports[graph.name].values():
            if len(report_list) > 0:
                has_report = True
        if not has_report:
            continue
        output[graph.name] = {}
        for vul in reports[graph.name].keys():
            if len(reports[graph.name][vul]) == 0:
                continue
            output[graph.name][vul] = {}
            for report in reports[graph.name][vul]:
                path = report.location.path
                line = report.location.start.line
                col = report.location.start.column
                path += f":{line}:{col}"
                if path not in output[graph.name][vul]:
                    output[graph.name][vul][path] = []
                if report.message not in output[graph.name][vul][path]:
                    output[graph.name][vul][path].append(report.message)
    with open(filename, "w", encoding="utf-8") as file:
        json.dump(output, file, indent=2)


@timeit
def detect(absolute_path):
    BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    sys.path.append(BASE_DIR)

    for sub in ["parser", "astgen", "typecheck", "utils", "cdggen", "detect", "target"]:
        path = os.path.join(BASE_DIR, "circheck", sub)
        if path not in sys.path:
            sys.path.append(path)

    ast = generate_ast(absolute_path)
    if ast is None:
        print("[Error]      Failed to generate AST.")
        return None, None
    print("[Success]    AST generated successfully.")

    checked, list_function, list_template = typecheck(ast)
    if checked is None:
        print("[Error]      Type checking failed.")
        return None, None
    print("[Success]    Type checking passed.")

    graphs = generate_cdg(ast, checked, list_function, list_template)
    if graphs is None:
        print("[Error]      Created CDG failed.")
        return None, None
    print("[Success]    CDG created successfully.")

    from Detect import Detector
    try:
        reports = Detector(graphs).detect()
        print("[Success]    Detection completed successfully.")
    except Exception as e:
        traceback.print_exc()
        print(f"[Error]      An error occurred during detection. {str(e)}")
        return None, None

    return graphs, reports



# Generated from circheck/parser/Circom.g4 by ANTLR 4.9.2
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .CircomParser import CircomParser
else:
    from CircomParser import CircomParser

# This class defines a complete generic visitor for a parse tree produced by CircomParser.

class CircomVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by CircomParser#program.
    def visitProgram(self, ctx:CircomParser.ProgramContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#version.
    def visitVersion(self, ctx:CircomParser.VersionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#main_option.
    def visitMain_option(self, ctx:CircomParser.Main_optionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#definition_block.
    def visitDefinition_block(self, ctx:CircomParser.Definition_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#definition_list.
    def visitDefinition_list(self, ctx:CircomParser.Definition_listContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#pragma_definition.
    def visitPragma_definition(self, ctx:CircomParser.Pragma_definitionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#custom_gate.
    def visitCustom_gate(self, ctx:CircomParser.Custom_gateContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#include_list.
    def visitInclude_list(self, ctx:CircomParser.Include_listContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#include_definition.
    def visitInclude_definition(self, ctx:CircomParser.Include_definitionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#public_list.
    def visitPublic_list(self, ctx:CircomParser.Public_listContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#main_component.
    def visitMain_component(self, ctx:CircomParser.Main_componentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#function_definition.
    def visitFunction_definition(self, ctx:CircomParser.Function_definitionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#template_definition.
    def visitTemplate_definition(self, ctx:CircomParser.Template_definitionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#identifier_list_option.
    def visitIdentifier_list_option(self, ctx:CircomParser.Identifier_list_optionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#custom_option.
    def visitCustom_option(self, ctx:CircomParser.Custom_optionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#parallel_option.
    def visitParallel_option(self, ctx:CircomParser.Parallel_optionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#block.
    def visitBlock(self, ctx:CircomParser.BlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#statement.
    def visitStatement(self, ctx:CircomParser.StatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#statement_list.
    def visitStatement_list(self, ctx:CircomParser.Statement_listContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#declaration_statement.
    def visitDeclaration_statement(self, ctx:CircomParser.Declaration_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#expression_statement.
    def visitExpression_statement(self, ctx:CircomParser.Expression_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#substitutions.
    def visitSubstitutions(self, ctx:CircomParser.SubstitutionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#substitutions_statement.
    def visitSubstitutions_statement(self, ctx:CircomParser.Substitutions_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#if_statement.
    def visitIf_statement(self, ctx:CircomParser.If_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#regular_statement.
    def visitRegular_statement(self, ctx:CircomParser.Regular_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#for_statement.
    def visitFor_statement(self, ctx:CircomParser.For_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#while_statement.
    def visitWhile_statement(self, ctx:CircomParser.While_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#equal_constraint_statement.
    def visitEqual_constraint_statement(self, ctx:CircomParser.Equal_constraint_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#return_statement.
    def visitReturn_statement(self, ctx:CircomParser.Return_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#assert_statement.
    def visitAssert_statement(self, ctx:CircomParser.Assert_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#log_statement.
    def visitLog_statement(self, ctx:CircomParser.Log_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#declaration.
    def visitDeclaration(self, ctx:CircomParser.DeclarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#identifier_list.
    def visitIdentifier_list(self, ctx:CircomParser.Identifier_listContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#tag_list.
    def visitTag_list(self, ctx:CircomParser.Tag_listContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#tuple_initialization.
    def visitTuple_initialization(self, ctx:CircomParser.Tuple_initializationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#simple_symbol.
    def visitSimple_symbol(self, ctx:CircomParser.Simple_symbolContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#simple_symbol_list.
    def visitSimple_symbol_list(self, ctx:CircomParser.Simple_symbol_listContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#complex_symbol.
    def visitComplex_symbol(self, ctx:CircomParser.Complex_symbolContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#some_symbol.
    def visitSome_symbol(self, ctx:CircomParser.Some_symbolContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#some_symbol_list.
    def visitSome_symbol_list(self, ctx:CircomParser.Some_symbol_listContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#var_decl.
    def visitVar_decl(self, ctx:CircomParser.Var_declContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#signal_decl.
    def visitSignal_decl(self, ctx:CircomParser.Signal_declContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#signal_header.
    def visitSignal_header(self, ctx:CircomParser.Signal_headerContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#signal_symbol.
    def visitSignal_symbol(self, ctx:CircomParser.Signal_symbolContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#signal_symbol_list.
    def visitSignal_symbol_list(self, ctx:CircomParser.Signal_symbol_listContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#component_decl.
    def visitComponent_decl(self, ctx:CircomParser.Component_declContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#var_access.
    def visitVar_access(self, ctx:CircomParser.Var_accessContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#var_access_list.
    def visitVar_access_list(self, ctx:CircomParser.Var_access_listContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#array_access.
    def visitArray_access(self, ctx:CircomParser.Array_accessContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#array_access_list.
    def visitArray_access_list(self, ctx:CircomParser.Array_access_listContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#component_access.
    def visitComponent_access(self, ctx:CircomParser.Component_accessContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#variable.
    def visitVariable(self, ctx:CircomParser.VariableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#expression.
    def visitExpression(self, ctx:CircomParser.ExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#expression1.
    def visitExpression1(self, ctx:CircomParser.Expression1Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#expression2.
    def visitExpression2(self, ctx:CircomParser.Expression2Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#expression3.
    def visitExpression3(self, ctx:CircomParser.Expression3Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#expression4.
    def visitExpression4(self, ctx:CircomParser.Expression4Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#expression5.
    def visitExpression5(self, ctx:CircomParser.Expression5Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#expression6.
    def visitExpression6(self, ctx:CircomParser.Expression6Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#expression7.
    def visitExpression7(self, ctx:CircomParser.Expression7Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#expression8.
    def visitExpression8(self, ctx:CircomParser.Expression8Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#expression9.
    def visitExpression9(self, ctx:CircomParser.Expression9Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#expression10.
    def visitExpression10(self, ctx:CircomParser.Expression10Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#expression11.
    def visitExpression11(self, ctx:CircomParser.Expression11Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#expression12.
    def visitExpression12(self, ctx:CircomParser.Expression12Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#expression13.
    def visitExpression13(self, ctx:CircomParser.Expression13Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#expression14.
    def visitExpression14(self, ctx:CircomParser.Expression14Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#twoElemsListable.
    def visitTwoElemsListable(self, ctx:CircomParser.TwoElemsListableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#log_arguement.
    def visitLog_arguement(self, ctx:CircomParser.Log_arguementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#log_list.
    def visitLog_list(self, ctx:CircomParser.Log_listContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#listable.
    def visitListable(self, ctx:CircomParser.ListableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#listable_prime.
    def visitListable_prime(self, ctx:CircomParser.Listable_primeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#listableAnon.
    def visitListableAnon(self, ctx:CircomParser.ListableAnonContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#listableAnon_prime.
    def visitListableAnon_prime(self, ctx:CircomParser.ListableAnon_primeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#listableWithInputNames.
    def visitListableWithInputNames(self, ctx:CircomParser.ListableWithInputNamesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#prefixOpcode.
    def visitPrefixOpcode(self, ctx:CircomParser.PrefixOpcodeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#compareOpcode.
    def visitCompareOpcode(self, ctx:CircomParser.CompareOpcodeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#shiftOpcode.
    def visitShiftOpcode(self, ctx:CircomParser.ShiftOpcodeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#add_sub_opcode.
    def visitAdd_sub_opcode(self, ctx:CircomParser.Add_sub_opcodeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#mul_div_opcode.
    def visitMul_div_opcode(self, ctx:CircomParser.Mul_div_opcodeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CircomParser#assign_opcode.
    def visitAssign_opcode(self, ctx:CircomParser.Assign_opcodeContext):
        return self.visitChildren(ctx)



del CircomParser


from __future__ import annotations
from dataclasses import dataclass, fields
from typing import Optional, List, Tuple
from abc import ABC, abstractmethod
from Visitor import Visitor
from antlr4 import Token


class AST(ABC):
    @abstractmethod
    def accept(self, v: Visitor, param):
        return v.visit(self, param)

    def __str__(self, indent: str = "", is_last: bool = True):
        label = self.__class__.__name__
        branch = "└── " if is_last else "├── "
        output = [f"{indent}{branch}{label}"]
        indent += "    " if is_last else "│   "

        for i, field in enumerate(fields(self)):
            value = getattr(self, field.name)
            is_last_field = i == len(fields(self)) - 1
            branch = "└── " if is_last_field else "├── "
            line = f"{indent}{branch}{field.name}:"

            if isinstance(value, AST):
                output.append(line)
                output.append(value.__str__(
                    indent + ("    " if is_last_field else "│   "), True))
            elif isinstance(value, list):
                output.append(line)
                for j, item in enumerate(value):
                    is_last_item = j == len(value) - 1
                    if isinstance(item, AST):
                        output.append(item.__str__(
                            indent + ("    " if is_last_field else "│   "), is_last_item))
                    else:
                        branch = "└── " if is_last_item else "├── "
                        output.append(f"{indent}    {branch}{item}")
            else:
                output.append(f"{line} {value}")

        return "\n".join(output)


class Expression(AST):
    pass


class Definition(AST):
    pass


class Statement(AST):
    pass


class VariableType(AST):
    pass


class Access(AST):
    pass


class LogArgument(AST):
    pass


@dataclass
class FileLocation(AST):
    path: str
    start: Token
    stop: Token

    def __str__(self, indent: str = "", is_last: bool = True):
        label = self.__class__.__name__
        branch = "└── " if is_last else "├── "
        output = [f"{indent}{branch}{label}"]
        indent += "    " if is_last else "│   "
        output.append(f"{indent}├── path: {self.path}")

        start_line = self.start.line
        start_column = self.start.column
        output.append(
            f"{indent}├── start: line {start_line}, column {start_column}")
        stop_line = self.stop.line
        stop_column = self.stop.column
        output.append(
            f"{indent}└── stop: line {stop_line}, column {stop_column}")

        return "\n".join(output)

    def accept(self, v: Visitor, param):
        return v.visitFileLocation(self, param)


@dataclass
class MainComponent(AST):
    locate: FileLocation
    publics: List[str]
    expr: Expression

    def accept(self, v: Visitor, param):
        return v.visitMainComponent(self, param)


@dataclass
class Include(AST):
    locate: FileLocation
    path: str

    def accept(self, v: Visitor, param):
        return v.visitInclude(self, param)


@dataclass
class Program(AST):
    locate: FileLocation
    compile_version: Optional[Tuple[int, int, int]]
    custom_gates: bool
    custom_gates_declare: bool
    includes: List[Include]
    definitions: List[Definition]
    main_component: Optional[MainComponent]

    def accept(self, v: Visitor, param):
        return v.visitProgram(self, param)


@dataclass
class Template(Definition):
    locate: FileLocation
    name_field: str
    args: List[str]
    body: Statement
    parallel: bool
    is_custom_gate: bool

    def accept(self, v: Visitor, param):
        return v.visitTemplate(self, param)


@dataclass
class Function(Definition):
    locate: FileLocation
    name_field: str
    args: List[str]
    body: Statement

    def accept(self, v: Visitor, param):
        return v.visitFunction(self, param)


@dataclass
class IfThenElse(Statement):
    locate: FileLocation
    cond: Expression
    if_case: Statement
    else_case: Optional[Statement]

    def accept(self, v: Visitor, param):
        return v.visitIfThenElse(self, param)


@dataclass
class While(Statement):
    locate: FileLocation
    cond: Expression
    stmt: Statement

    def accept(self, v: Visitor, param):
        return v.visitWhile(self, param)


@dataclass
class Return(Statement):
    locate: FileLocation
    value: Expression

    def accept(self, v: Visitor, param):
        return v.visitReturn(self, param)


@dataclass
class InitializationBlock(Statement):
    locate: FileLocation
    xtype: VariableType
    initializations: List[Statement]

    def accept(self, v: Visitor, param):
        return v.visitInitializationBlock(self, param)


@dataclass
class Declaration(Statement):
    locate: FileLocation
    xtype: VariableType
    name: str
    dimensions: List[Expression]
    is_constant: bool

    def accept(self, v: Visitor, param):
        return v.visitDeclaration(self, param)


@dataclass
class Substitution(Statement):
    locate: FileLocation
    var: str
    access: List[Access]
    op: str
    rhe: Expression

    def accept(self, v: Visitor, param):
        return v.visitSubstitution(self, param)


@dataclass
class MultiSubstitution(Statement):
    locate: FileLocation
    lhe: Expression
    op: str
    rhe: Expression

    def accept(self, v: Visitor, param):
        return v.visitMultiSubstitution(self, param)


@dataclass
class ConstraintEquality(Statement):
    locate: FileLocation
    lhe: Expression
    rhe: Expression

    def accept(self, v: Visitor, param):
        return v.visitConstraintEquality(self, param)


@dataclass
class LogCall(Statement):
    locate: FileLocation
    args: List[LogArgument]

    def accept(self, v: Visitor, param):
        return v.visitLogCall(self, param)


@dataclass
class Block(Statement):
    locate: FileLocation
    stmts: List[Statement]

    def accept(self, v: Visitor, param):
        return v.visitBlock(self, param)


@dataclass
class Assert(Statement):
    locate: FileLocation
    arg: Expression

    def accept(self, v: Visitor, param):
        return v.visitAssert(self, param)


@dataclass
class Var(VariableType):
    def accept(self, v: Visitor, param):
        return v.visitVar(self, param)


@dataclass
class Signal(VariableType):
    locate: FileLocation
    signal_type: str
    tag_list: List[str]

    def accept(self, v: Visitor, param):
        return v.visitSignal(self, param)


@dataclass
class Component(VariableType):
    def accept(self, v: Visitor, param):
        return v.visitComponent(self, param)


@dataclass
class AnonymousComponent(VariableType):
    def accept(self, v: Visitor, param):
        return v.visitAnonymousComponent(self, param)


@dataclass
class InfixOp(Expression):
    locate: FileLocation
    lhe: Expression
    infix_op: str
    rhe: Expression

    def accept(self, v: Visitor, param):
        return v.visitInfixOp(self, param)


@dataclass
class PrefixOp(Expression):
    locate: FileLocation
    prefix_op: str
    rhe: Expression

    def accept(self, v: Visitor, param):
        return v.visitPrefixOp(self, param)


@dataclass
class InlineSwitchOp(Expression):
    locate: FileLocation
    cond: Expression
    if_true: Expression
    if_false: Expression

    def accept(self, v: Visitor, param):
        return v.visitInlineSwitchOp(self, param)


@dataclass
class ParallelOp(Expression):
    locate: FileLocation
    rhe: Expression

    def accept(self, v: Visitor, param):
        return v.visitParallelOp(self, param)


@dataclass
class Variable(Expression):
    locate: FileLocation
    name: str
    access: List[Access]

    def accept(self, v: Visitor, param):
        return v.visitVariable(self, param)


@dataclass
class Number(Expression):
    locate: FileLocation
    value: int

    def accept(self, v: Visitor, param):
        return v.visitNumber(self, param)


@dataclass
class Call(Expression):
    locate: FileLocation
    id: str
    args: List[Expression]

    def accept(self, v: Visitor, param):
        return v.visitCall(self, param)


@dataclass
class AnonymousComponentExpr(Expression):
    locate: FileLocation
    id: str
    is_parallel: bool
    params: List[Expression]
    signals: List[Expression]
    names: Optional[List[Tuple[str, str]]]

    def accept(self, v: Visitor, param):
        return v.visitAnonymousComponentExpr(self, param)


@dataclass
class ArrayInLine(Expression):
    locate: FileLocation
    values: List[Expression]

    def accept(self, v: Visitor, param):
        return v.visitArrayInLine(self, param)


@dataclass
class TupleExpr(Expression):
    locate: FileLocation
    values: List[Expression]

    def accept(self, v: Visitor, param):
        return v.visitTupleExpr(self, param)


@dataclass
class ComponentAccess(Access):
    locate: FileLocation
    name: str

    def accept(self, v: Visitor, param):
        return v.visitComponentAccess(self, param)


@dataclass
class ArrayAccess(Access):
    locate: FileLocation
    expr: Expression

    def accept(self, v: Visitor, param):
        return v.visitArrayAccess(self, param)


@dataclass
class LogStr(LogArgument):
    locate: FileLocation
    value: str

    def accept(self, v: Visitor, param):
        return v.visitLogStr(self, param)


@dataclass
class LogExp(LogArgument):
    locate: FileLocation
    expr: Expression

    def accept(self, v: Visitor, param):
        return v.visitLogExp(self, param)



from enum import Enum
import colorama
from AST import FileLocation

colorama.init(autoreset=True)


class ReportType(Enum):
    ERROR = "Error"
    WARNING = "Warning"


class Report(Exception):
    def __init__(self, type_: ReportType, location: FileLocation, message: str, has_location: bool = True):
        self.type = type_
        self.location = location
        self.message = message
        self.has_location = has_location

    def __str__(self):
        color = colorama.Fore.RED if self.type == ReportType.ERROR else colorama.Fore.YELLOW
        prefix = f"{color}[{self.type.value}]{colorama.Style.RESET_ALL}"
        path = self.location.path
        if self.has_location:
            line = self.location.start.line
            col = self.location.start.column
            location_str = f"{path}:{line}:{col}"
        else:
            location_str = path
        return f"{prefix}    In {location_str}\n{' ' * 13}{self.message}"

    def show(self):
        print(self.__str__())



from enum import Enum, auto
from StaticCheck import SignalType
from tqdm import tqdm


class NodeType(Enum):
    SIGNAL = auto()
    CONSTANT = auto()


class EdgeType(Enum):
    DEPEND = auto()
    CONSTRAINT = auto()


class Node:
    def __init__(self, locate, id, node_type, signal_type, component):
        self.locate = locate
        self.id = id
        self.node_type = node_type
        self.signal_type = signal_type
        self.component = component
        # node == edge.node_from
        self.flow_to = []
        # node = edge.node_to
        self.flow_from = []

    def __repr__(self):
        return f"Node({self.id})"

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other):
        return isinstance(other, Node) and self.id == other.id

    def is_signal(self):
        return self.node_type == NodeType.SIGNAL

    def is_signal_in(self):
        return self.is_signal() and self.signal_type == SignalType.INPUT

    def is_signal_out(self):
        return self.is_signal() and self.signal_type == SignalType.OUTPUT

    def is_signal_in_of(self, component):
        return self.is_signal_in() and self.component == component

    def is_signal_out_of(self, component):
        return self.is_signal_out() and self.component == component

    def is_signal_of(self, component):
        return self.is_signal() and self.component == component


class Edge:
    def __init__(self, node_from, node_to, edge_type, name, ast=None):
        self.node_from = node_from
        self.node_to = node_to
        self.edge_type = edge_type
        self.name = name
        self.ast = ast


def getEdgeName(edge_type, nFrom, nTo):
    if edge_type == EdgeType.DEPEND:
        return f"data:{nFrom.id}-{nTo.id}"
    else:
        return f"constraint:{nFrom.id}-{nTo.id}"


class CircuitDependenceGraph:
    def __init__(self, edges, nodes, name, components):
        self.edges = edges
        self.nodes = nodes
        self.name = name
        self.components = components
        self.node_flows_to = {}

    def is_signal(self, node):
        return node.node_type == NodeType.SIGNAL

    def is_signal_in(self, node):
        return self.is_signal(node) and node.signal_type == SignalType.INPUT

    def is_signal_out(self, node):
        return self.is_signal(node) and node.signal_type == SignalType.OUTPUT

    def is_signal_of(self, node, component):
        return self.is_signal(node) and node.component == component

    def is_signal_in_of(self, node, component):
        return self.is_signal_of(node, component) and self.is_signal_in(node)

    def is_signal_out_of(self, node, component):
        return self.is_signal_of(node, component) and self.is_signal_out(node)

    def has_path_depend(self, a: Node, b: Node) -> bool:
        visited = set()
        stack = [a]
        while stack:
            current = stack.pop()
            if current.id == b.id:
                return True
            if current.id in visited:
                continue
            visited.add(current.id)
            for edge in a.flow_to:
                if edge.edge_type == EdgeType.DEPEND:
                    stack.append(edge.node_to)
        return False

    def has_path_constraint(self, a: Node, b: Node) -> bool:
        visited = set()
        stack = [a]
        while stack:
            current = stack.pop()
            if current.id == b.id:
                return True
            if current.id in visited:
                continue
            visited.add(current.id)
            for edge in current.flow_to:
                if edge.edge_type == EdgeType.CONSTRAINT:
                    stack.append(edge.node_to)
        return False

    def build_conditional_depend_edges(self, graphs):
        print(
            f"[Info]       Building conditional dependency edges of {self.name}...")
        for u in tqdm(self.nodes.values()):
            if self.is_signal_in(u):
                component = u.component
                for v_id in self.components[component][SignalType.OUTPUT]:
                    v = self.nodes[v_id]
                    if self.has_path_depend(u, v):
                        edge_name = getEdgeName(EdgeType.DEPEND, u, v)
                        if edge_name not in self.edges:
                            edge = self.edges[edge_name] = Edge(
                                u, v, EdgeType.DEPEND, edge_name)
                            u.flow_to.append(edge)
                            v.flow_from.append(edge)
                    else:
                        graph_name = component.split("|")[0]
                        if graph_name != self.name:
                            graph = graphs[graph_name]
                            a_id = u.id.split(".")[1]
                            b_id = v_id.split(".")[1]
                            if a_id not in graph.nodes or b_id not in graph.nodes:
                                continue
                            node_a = graph.nodes[a_id]
                            node_b = graph.nodes[b_id]
                            if graph.has_path_depend(node_a, node_b):
                                edge_name = getEdgeName(EdgeType.DEPEND, u, v)
                                if edge_name not in self.edges:
                                    edge = self.edges[edge_name] = Edge(
                                        u, v, EdgeType.DEPEND, edge_name)
                                    u.flow_to.append(edge)
                                    v.flow_from.append(edge)

    def build_condition_constraint_edges(self, graphs):
        print(
            f"[Info]       Building condition constraint edges of {self.name}...")
        for u in tqdm(self.nodes.values()):
            if self.is_signal_in(u):
                component = u.component
                for v_id in self.components[component][SignalType.OUTPUT]:
                    v = self.nodes[v_id]
                    if self.has_path_constraint(u, v):
                        edge_name = getEdgeName(EdgeType.CONSTRAINT, u, v)
                        if edge_name not in self.edges:
                            edge = self.edges[edge_name] = Edge(
                                u, v, EdgeType.CONSTRAINT, edge_name)
                            u.flow_to.append(edge)
                            v.flow_from.append(edge)
                            edge_name_reversed = getEdgeName(
                                EdgeType.CONSTRAINT, v, u)
                            if edge_name_reversed not in self.edges:
                                edge = self.edges[edge_name] = Edge(
                                    v, u, EdgeType.CONSTRAINT, edge_name_reversed)
                                v.flow_to.append(edge)
                                u.flow_from.append(edge)
                    else:
                        graph_name = component.split("|")[0]
                        if graph_name != self.name:
                            graph = graphs[graph_name]
                            a_id = u.id.split(".")[1]
                            b_id = v_id.split(".")[1]
                            if a_id not in graph.nodes or b_id not in graph.nodes:
                                continue
                            node_a = graph.nodes[a_id]
                            node_b = graph.nodes[b_id]
                            if graph.has_path_constraint(node_a, node_b):
                                edge_name = getEdgeName(
                                    EdgeType.CONSTRAINT, u, v)
                                if edge_name not in self.edges:
                                    edge = self.edges[edge_name] = Edge(
                                        u, v, EdgeType.CONSTRAINT, edge_name)
                                    u.flow_to.append(edge)
                                    v.flow_from.append(edge)
                                    edge_name_reversed = getEdgeName(
                                        EdgeType.CONSTRAINT, v, u)
                                    if edge_name_reversed not in self.edges:
                                        edge = self.edges[edge_name] = Edge(
                                            v, u, EdgeType.CONSTRAINT, edge_name_reversed)
                                        v.flow_to.append(edge)
                                        u.flow_from.append(edge)

    def flows_to(self, node):
        if node.id in self.node_flows_to:
            return self.node_flows_to[node.id]
        flows_to_set = set()
        for edge in node.flow_to:
            if edge.edge_type == EdgeType.CONSTRAINT:
                continue
            node_to = edge.node_to
            if node_to.is_signal():
                flows_to_sub = self.flows_to(node_to)
                flows_to_set.union(flows_to_sub)
                flows_to_set.add(node_to.id)
            else:
                for e1 in node_to.flow_to:
                    if e1.edge_type == EdgeType.CONSTRAINT:
                        continue
                    to_1 = e1.node_to
                    if to_1.is_signal():
                        flows_to_sub = self.flows_to(to_1)
                        flows_to_set.union(flows_to_sub)
                        flows_to_set.add(to_1.id)
        self.node_flows_to[node.id] = flows_to_set
        return flows_to_set

    def compute(self):
        for node in self.nodes.values():
            if node.is_signal():
                self.flows_to(node)

    def build_graph(self, graphs):
        self.build_conditional_depend_edges(graphs)
        self.build_condition_constraint_edges(graphs)
        self.compute()



from setuptools import setup, find_packages

setup(
    name="circheck",
    version="0.1.0",
    description="Static analysis tool to detect ZKP vulnerabilities in Circom circuits.",
    author="Dang Duong Minh Nhat",
    author_email="dangduongminhnhat2003@gmail.com",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    install_requires=[
        "antlr4-python3-runtime"
    ],
    entry_points={
        'console_scripts': [
            'circheck = circheck.cli:main',
        ],
    },
)
